# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	 http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ######################################################################
# 형태소분석 기반 BERT 모델 MRC Fine-tuning 샘플
# (original: Hugging-face BERT example code)
# 수정: joonho.lim
# 일자: 2019-05-27
#
"""Run BERT on SQuAD."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import collections
import logging
import json
import math
import os
import random
import pickle
from tqdm import tqdm, trange

import numpy as np
import torch
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, SequentialSampler
from torch.utils.data.distributed import DistributedSampler

############################################
### joonho.lim @ 2019-03-15
from pytorch_pretrained_bert import BasicTokenizer, BertTokenizer
from pytorch_pretrained_bert.modeling import BertForQuestionAnswering
from pytorch_pretrained_bert.optimization import BertAdam
from pytorch_pretrained_bert.file_utils import PYTORCH_PRETRAINED_BERT_CACHE

import time
import urllib3

logging.basicConfig(format = '%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
					datefmt = '%m/%d/%Y %H:%M:%S',
					level = logging.INFO)
logger = logging.getLogger(__name__)


class SquadExample(object):
	"""A single training/test example for the Squad dataset."""

	def __init__(self,
				 qas_id,
				 q_raw_text,
				 q_morp_token,
				 p_raw_text,
				 p_morp_token,
				 p_morp_position_list,
				 a_raw_text=None,
				 a_morp_token=None,
				 a_begin_morp=None,
				 a_end_morp=None):
		self.qas_id = qas_id
		self.q_raw_text = q_raw_text
		self.q_morp_token = q_morp_token
		self.p_raw_text = p_raw_text
		self.p_morp_token = p_morp_token
		self.p_morp_position_list = p_morp_position_list
		self.a_raw_text = a_raw_text
		self.a_morp_token = a_morp_token
		self.a_begin_morp = a_begin_morp
		self.a_end_morp = a_end_morp
		
		self.p_raw_bytes = p_raw_text.encode()
		self.p_morp_position_list.append ( len(self.p_raw_bytes) )
		
		##########################################################
		### joonho.lim @ 2019-03-15
		### check difference answer span between answer raw text and morphology anlysis boudary 
		if a_raw_text!= None and a_end_morp != None and len(p_morp_position_list) > a_end_morp :
			begin_pos = p_morp_position_list[a_begin_morp]
			end_pos = p_morp_position_list[a_end_morp+1]
			pred_answer = self.p_raw_bytes[begin_pos:end_pos].decode().strip()
			if self.a_raw_text != pred_answer :
				logger.info ( "[diff answer span] %s\t%s" % (self.a_raw_text, pred_answer) )

	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		s = ""
		s += "qas_id: %s" % (self.qas_id)
		s += ", question_text: %s" % (self.q_raw_text)
		s += ", doc_text: [%s]" % (self.p_raw_text)
		if self.start_position:	s += ", start_position: %d" % (self.a_begin_morp)
		if self.start_position:	s += ", end_position: %d" % (self.a_end_morp)
		return s




class InputFeatures(object):
	"""A single set of features of data."""

	def __init__(self,
				 unique_id,
				 example_index,
				 doc_span_index,
				 tokens,
				 token_to_orig_map,
				 token_is_max_context,
				 input_ids,
				 input_mask,
				 segment_ids,
				 start_position=None,
				 end_position=None):
		self.unique_id = unique_id
		self.example_index = example_index
		self.doc_span_index = doc_span_index
		self.tokens = tokens
		self.token_to_orig_map = token_to_orig_map
		self.token_is_max_context = token_is_max_context
		self.input_ids = input_ids
		self.input_mask = input_mask
		self.segment_ids = segment_ids
		self.start_position = start_position
		self.end_position = end_position

		
##########################################################
### joonho.lim @ 2019-03-15
### convert language anlysis result to inner data structure
def represent_ndoc ( ndoc ) :
	text = ''
	morp_list = []
	position_list = []
	
	for sentence in ndoc['sentence'] :
		text += sentence['text']
		begin_morp_id = len(morp_list)
		
		for morp in sentence['morp'] :
			morp_list.append( morp['lemma'] + '/' + morp['type'] )
			position_list.append( int(morp['position']) )
	
	return { 'text': text, 'morp_list':morp_list, 'position_list':position_list }
	

		
##########################################################
### joonho.lim @ 2019-03-15
### do morphology analysis using OpenAPI service
def do_lang ( openapi_key, text ) :
	openApiURL = "http://aiopen.etri.re.kr:8000/WiseNLU"
	 
	requestJson = { "access_key": openapi_key, "argument": { "text": text, "analysis_code": "morp" } }
	 
	http = urllib3.PoolManager()
	response = http.request( "POST", openApiURL, headers={"Content-Type": "application/json; charset=UTF-8"}, body=json.dumps(requestJson) )

	return response.data.decode()


##########################################################
### joonho.lim @ 2019-03-15
### read squad example file and do morphology analysis
def read_squad_examples_and_do_lang(input_file, openapi_key, is_training):
	"""Read a SQuAD json file into a list of SquadExample."""
	if os.path.isfile(input_file) == False :
		raise ValueError("not exist file or folder : %s" % input_file)
		
	with open(input_file, "r", encoding='utf-8') as reader:
		input_data = json.load(reader)
	
	pqa_list = []
	for paragraphs_title in input_data['data'] :
		for pq in paragraphs_title['paragraphs'] :
			passage_text = pq['context']
			passage_lang = do_lang( openapi_key, passage_text )
			p_json = json.loads(passage_lang)['return_object']
			rep_p = represent_ndoc(p_json)
			
			for qa in pq['qas'] :
				qas_id = qa['id']
				question_text = qa['question']
				question_lang = do_lang( openapi_key, question_text )
				q_json = json.loads(question_lang)['return_object']
				rep_q = represent_ndoc(q_json)
				
				rep_a = {}
				if is_training :
					print('')
					#rep_a = mapping_answer_korquad( p_json, rep_p, qa['answers'][0]['text'], qa['answers'][0]['answer_start'] )
			
				pqa_list.append( {'id':qas_id, 'passage':rep_p, 'question':rep_q, 'answer':rep_a} )
	
	return read_squad_examples(pqa_list, is_training)


def read_squad_examples(input_data, is_training):
	"""Read a SQuAD json file into a list of SquadExample."""
	examples = []
	for pqa in input_data :
		a_raw_text = None
		a_morp_token = None
		a_begin_morp = None
		a_end_morp = None
		if is_training:
			a_raw_text = pqa['answer']['text']
			a_begin_morp = pqa['answer']['begin_morp']
			a_end_morp = pqa['answer']['end_morp']
			a_morp_token = pqa['passage']['morp_list'][a_begin_morp : a_end_morp+1]
			
		example = SquadExample(
			qas_id = pqa['id'],
			q_raw_text = pqa['question']['text'],
			q_morp_token = pqa['question']['morp_list'],
			p_raw_text = pqa['passage']['text'],
			p_morp_token = pqa['passage']['morp_list'],
			p_morp_position_list = pqa['passage']['position_list'],
			a_raw_text = a_raw_text,
			a_morp_token = a_morp_token,
			a_begin_morp = a_begin_morp,
			a_end_morp = a_end_morp)
		examples.append(example)
		
	logger.info( 'len(examples) : %d'% len(examples) )
	return examples		
		

def convert_examples_to_features(examples, tokenizer, max_seq_length,
								 doc_stride, max_query_length, is_training, verbose=True):
	"""Loads a data file into a list of `InputBatch`s."""

	unique_id = 1000000000

	features = []
	for (example_index, example) in enumerate(examples):
		query_tokens = []
		for q_morp in example.q_morp_token :
			query_tokens.extend( tokenizer.tokenize(q_morp) )
		if len(query_tokens) > max_query_length:
			query_tokens = query_tokens[0:max_query_length]
		
		tok_to_orig_index = []
		orig_to_tok_index = []
		all_doc_tokens = []
		for (i, token) in enumerate(example.p_morp_token):
			orig_to_tok_index.append(len(all_doc_tokens))
			sub_tokens = tokenizer.tokenize(token)
			for sub_token in sub_tokens:
				tok_to_orig_index.append(i)
				all_doc_tokens.append(sub_token)

		tok_start_position = None
		tok_end_position = None
		if is_training:
			tok_start_position = orig_to_tok_index[ example.a_begin_morp ]
			if example.a_end_morp+1 < len(orig_to_tok_index) :
				tok_end_position = orig_to_tok_index[ example.a_end_morp+1 ] - 1
			else :
				tok_end_position = orig_to_tok_index[ -1 ]

				
		# The -3 accounts for [CLS], [SEP] and [SEP]
		max_tokens_for_doc = max_seq_length - len(query_tokens) - 3

		# We can have documents that are longer than the maximum sequence length.
		# To deal with this we do a sliding window approach, where we take chunks
		# of the up to our max length with a stride of `doc_stride`.
		_DocSpan = collections.namedtuple(  # pylint: disable=invalid-name
			"DocSpan", ["start", "length"])
		doc_spans = []
		start_offset = 0
		while start_offset < len(all_doc_tokens):
			length = len(all_doc_tokens) - start_offset
			if length > max_tokens_for_doc:
				length = max_tokens_for_doc
			doc_spans.append(_DocSpan(start=start_offset, length=length))
			if start_offset + length == len(all_doc_tokens):
				break
			start_offset += min(length, doc_stride)

		for (doc_span_index, doc_span) in enumerate(doc_spans):
			tokens = []
			token_to_orig_map = {}
			token_is_max_context = {}
			segment_ids = []
			tokens.append("[CLS]")
			segment_ids.append(0)
			for token in query_tokens:
				tokens.append(token)
				segment_ids.append(0)
			tokens.append("[SEP]")
			segment_ids.append(0)

			for i in range(doc_span.length):
				split_token_index = doc_span.start + i
				token_to_orig_map[len(tokens)] = tok_to_orig_index[split_token_index]

				is_max_context = _check_is_max_context(doc_spans, doc_span_index,
													   split_token_index)
				token_is_max_context[len(tokens)] = is_max_context
				tokens.append(all_doc_tokens[split_token_index])
				segment_ids.append(1)
			tokens.append("[SEP]")
			segment_ids.append(1)

			input_ids = tokenizer.convert_tokens_to_ids(tokens)

			# The mask has 1 for real tokens and 0 for padding tokens. Only real
			# tokens are attended to.
			input_mask = [1] * len(input_ids)

			# Zero-pad up to the sequence length.
			while len(input_ids) < max_seq_length:
				input_ids.append(0)
				input_mask.append(0)
				segment_ids.append(0)

			assert len(input_ids) == max_seq_length
			assert len(input_mask) == max_seq_length
			assert len(segment_ids) == max_seq_length

			start_position = None
			end_position = None
			if is_training:
				# For training, if our document chunk does not contain an annotation
				# we throw it out, since there is nothing to predict.
				doc_start = doc_span.start
				doc_end = doc_span.start + doc_span.length - 1
				if (example.a_begin_morp < doc_start or
						example.a_end_morp < doc_start or
						example.a_begin_morp > doc_end or example.a_end_morp > doc_end):
					continue

				doc_offset = len(query_tokens) + 2
				start_position = tok_start_position - doc_start + doc_offset
				end_position = tok_end_position - doc_start + doc_offset

			if verbose == True and example_index < 10:
				logger.info("*** Example ***")
				logger.info("unique_id: %s" % (unique_id))
				logger.info("example_index: %s" % (example_index))
				logger.info("doc_span_index: %s" % (doc_span_index))
				logger.info("tokens: %s" % " ".join(tokens))
				logger.info("token_to_orig_map: %s" % " ".join([
					"%d:%d" % (x, y) for (x, y) in token_to_orig_map.items()]))
				logger.info("token_is_max_context: %s" % " ".join([
					"%d:%s" % (x, y) for (x, y) in token_is_max_context.items()
				]))
				logger.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
				logger.info(
					"input_mask: %s" % " ".join([str(x) for x in input_mask]))
				logger.info(
					"segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
				if is_training:
					answer_text = " ".join(tokens[start_position:(end_position + 1)])
					logger.info("start_position: %d" % (start_position))
					logger.info("end_position: %d" % (end_position))
					logger.info("answer: %s" % (answer_text))
					logger.info("orig_answer: %s" % (example.a_raw_text))

			features.append(
				InputFeatures(
					unique_id=unique_id,
					example_index=example_index,
					doc_span_index=doc_span_index,
					tokens=tokens,
					token_to_orig_map=token_to_orig_map,
					token_is_max_context=token_is_max_context,
					input_ids=input_ids,
					input_mask=input_mask,
					segment_ids=segment_ids,
					start_position=start_position,
					end_position=end_position))
			unique_id += 1

	return features



def _check_is_max_context(doc_spans, cur_span_index, position):
	"""Check if this is the 'max context' doc span for the token."""

	# Because of the sliding window approach taken to scoring documents, a single
	# token can appear in multiple documents. E.g.
	#  Doc: the man went to the store and bought a gallon of milk
	#  Span A: the man went to the
	#  Span B: to the store and bought
	#  Span C: and bought a gallon of
	#  ...
	#
	# Now the word 'bought' will have two scores from spans B and C. We only
	# want to consider the score with "maximum context", which we define as
	# the *minimum* of its left and right context (the *sum* of left and
	# right context will always be the same, of course).
	#
	# In the example the maximum context for 'bought' would be span C since
	# it has 1 left context and 3 right context, while span B has 4 left context
	# and 0 right context.
	best_score = None
	best_span_index = None
	for (span_index, doc_span) in enumerate(doc_spans):
		end = doc_span.start + doc_span.length - 1
		if position < doc_span.start:
			continue
		if position > end:
			continue
		num_left_context = position - doc_span.start
		num_right_context = end - position
		score = min(num_left_context, num_right_context) + 0.01 * doc_span.length
		if best_score is None or score > best_score:
			best_score = score
			best_span_index = span_index

	return cur_span_index == best_span_index



RawResult = collections.namedtuple("RawResult",
								   ["unique_id", "start_logits", "end_logits"])


def write_predictions(all_examples, all_features, all_results, n_best_size,
					  max_answer_length, do_lower_case, output_prediction_file,
					  output_nbest_file, verbose_logging):
	"""Write final predictions to the json file."""
	logger.info("Writing predictions to: %s" % (output_prediction_file))
	logger.info("Writing nbest to: %s" % (output_nbest_file))

	(all_predictions, all_nbest_json) = get_predictions( 
					all_examples, all_features, all_results, n_best_size,
					  max_answer_length, do_lower_case, verbose_logging )
					  
	with open(output_prediction_file, "w") as writer:
		writer.write(json.dumps(all_predictions, indent=4) + "\n")

	with open(output_nbest_file, "w") as writer:
		writer.write(json.dumps(all_nbest_json, indent=4) + "\n")

		
def get_predictions(all_examples, all_features, all_results, n_best_size,
					  max_answer_length, do_lower_case, verbose_logging):

	example_index_to_features = collections.defaultdict(list)
	for feature in all_features:
		example_index_to_features[feature.example_index].append(feature)

	unique_id_to_result = {}
	for result in all_results:
		unique_id_to_result[result.unique_id] = result

	_PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
		"PrelimPrediction",
		["feature_index", "start_index", "end_index", "start_logit", "end_logit"])

	all_predictions = collections.OrderedDict()
	all_nbest_json = collections.OrderedDict()
	for (example_index, example) in enumerate(all_examples):
		features = example_index_to_features[example_index]

		prelim_predictions = []
		for (feature_index, feature) in enumerate(features):
			result = unique_id_to_result[feature.unique_id]

			start_indexes = _get_best_indexes(result.start_logits, n_best_size)
			end_indexes = _get_best_indexes(result.end_logits, n_best_size)
			for start_index in start_indexes:
				for end_index in end_indexes:
					# We could hypothetically create invalid predictions, e.g., predict
					# that the start of the span is in the question. We throw out all
					# invalid predictions.
					if start_index >= len(feature.tokens):
						continue
					if end_index >= len(feature.tokens):
						continue
					if start_index not in feature.token_to_orig_map:
						continue
					if end_index not in feature.token_to_orig_map:
						continue
					if not feature.token_is_max_context.get(start_index, False):
						continue
					if end_index < start_index:
						continue
					length = end_index - start_index + 1
					if length > max_answer_length:
						continue
					prelim_predictions.append(
						_PrelimPrediction(
							feature_index=feature_index,
							start_index=start_index,
							end_index=end_index,
							start_logit=result.start_logits[start_index],
							end_logit=result.end_logits[end_index]))

		prelim_predictions = sorted(
			prelim_predictions,
			key=lambda x: (x.start_logit + x.end_logit),
			reverse=True)

		_NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
			"NbestPrediction", ["text", "start_logit", "end_logit"])

		seen_predictions = {}
		nbest = []
		for pred in prelim_predictions:
			if len(nbest) >= n_best_size:
				break
			feature = features[pred.feature_index]

			##########################################################
			### joonho.lim @ 2019-03-15
			### extract final text
			tok_tokens = feature.tokens[pred.start_index:(pred.end_index + 1)]
			orig_doc_start = feature.token_to_orig_map[pred.start_index]
			orig_doc_end = feature.token_to_orig_map[pred.end_index]
			
			p_begin_morp = example.p_morp_position_list[ orig_doc_start ]
			p_end_morp = example.p_morp_position_list[ orig_doc_end+1 ]
			
			final_text = example.p_raw_bytes[ p_begin_morp:p_end_morp ].decode().strip()
			
			if final_text in seen_predictions:
				continue

			seen_predictions[final_text] = True
			nbest.append(
				_NbestPrediction(
					text=final_text,
					start_logit=pred.start_logit,
					end_logit=pred.end_logit))

		# In very rare edge cases we could have no valid predictions. So we
		# just create a nonce prediction in this case to avoid failure.
		if not nbest:
			nbest.append(
				_NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0))

		assert len(nbest) >= 1

		total_scores = []
		for entry in nbest:
			total_scores.append(entry.start_logit + entry.end_logit)

		probs = _compute_softmax(total_scores)

		nbest_json = []
		for (i, entry) in enumerate(nbest):
			output = collections.OrderedDict()
			output["text"] = entry.text
			output["probability"] = probs[i]
			output["start_logit"] = entry.start_logit
			output["end_logit"] = entry.end_logit
			nbest_json.append(output)

		assert len(nbest_json) >= 1

		all_predictions[example.qas_id] = nbest_json[0]["text"]
		all_nbest_json[example.qas_id] = nbest_json

	return (all_predictions, all_nbest_json)

	
def get_final_text(pred_text, orig_text, do_lower_case, verbose_logging=False):
	"""Project the tokenized prediction back to the original text."""

	# When we created the data, we kept track of the alignment between original
	# (whitespace tokenized) tokens and our WordPiece tokenized tokens. So
	# now `orig_text` contains the span of our original text corresponding to the
	# span that we predicted.
	#
	# However, `orig_text` may contain extra characters that we don't want in
	# our prediction.
	#
	# For example, let's say:
	#   pred_text = steve smith
	#   orig_text = Steve Smith's
	#
	# We don't want to return `orig_text` because it contains the extra "'s".
	#
	# We don't want to return `pred_text` because it's already been normalized
	# (the SQuAD eval script also does punctuation stripping/lower casing but
	# our tokenizer does additional normalization like stripping accent
	# characters).
	#
	# What we really want to return is "Steve Smith".
	#
	# Therefore, we have to apply a semi-complicated alignment heruistic between
	# `pred_text` and `orig_text` to get a character-to-charcter alignment. This
	# can fail in certain cases in which case we just return `orig_text`.

	def _strip_spaces(text):
		ns_chars = []
		ns_to_s_map = collections.OrderedDict()
		for (i, c) in enumerate(text):
			if c == " ":
				continue
			ns_to_s_map[len(ns_chars)] = i
			ns_chars.append(c)
		ns_text = "".join(ns_chars)
		return (ns_text, ns_to_s_map)

	# We first tokenize `orig_text`, strip whitespace from the result
	# and `pred_text`, and check if they are the same length. If they are
	# NOT the same length, the heuristic has failed. If they are the same
	# length, we assume the characters are one-to-one aligned.
	tokenizer = BasicTokenizer(do_lower_case=do_lower_case)

	tok_text = " ".join(tokenizer.tokenize(orig_text))

	start_position = tok_text.find(pred_text)
	if start_position == -1:
		if verbose_logging:
			logger.info(
				"Unable to find text: '%s' in '%s'" % (pred_text, orig_text))
		return orig_text
	end_position = start_position + len(pred_text) - 1

	(orig_ns_text, orig_ns_to_s_map) = _strip_spaces(orig_text)
	(tok_ns_text, tok_ns_to_s_map) = _strip_spaces(tok_text)

	if len(orig_ns_text) != len(tok_ns_text):
		if verbose_logging:
			logger.info("Length not equal after stripping spaces: '%s' vs '%s'",
							orig_ns_text, tok_ns_text)
		return orig_text

	# We then project the characters in `pred_text` back to `orig_text` using
	# the character-to-character alignment.
	tok_s_to_ns_map = {}
	for (i, tok_index) in tok_ns_to_s_map.items():
		tok_s_to_ns_map[tok_index] = i

	orig_start_position = None
	if start_position in tok_s_to_ns_map:
		ns_start_position = tok_s_to_ns_map[start_position]
		if ns_start_position in orig_ns_to_s_map:
			orig_start_position = orig_ns_to_s_map[ns_start_position]

	if orig_start_position is None:
		if verbose_logging:
			logger.info("Couldn't map start position")
		return orig_text

	orig_end_position = None
	if end_position in tok_s_to_ns_map:
		ns_end_position = tok_s_to_ns_map[end_position]
		if ns_end_position in orig_ns_to_s_map:
			orig_end_position = orig_ns_to_s_map[ns_end_position]

	if orig_end_position is None:
		if verbose_logging:
			logger.info("Couldn't map end position")
		return orig_text

	output_text = orig_text[orig_start_position:(orig_end_position + 1)]
	return output_text


def _get_best_indexes(logits, n_best_size):
	"""Get the n-best logits from a list."""
	index_and_score = sorted(enumerate(logits), key=lambda x: x[1], reverse=True)

	best_indexes = []
	for i in range(len(index_and_score)):
		if i >= n_best_size:
			break
		best_indexes.append(index_and_score[i][0])
	return best_indexes


def _compute_softmax(scores):
	"""Compute softmax probability over raw logits."""
	if not scores:
		return []

	max_score = None
	for score in scores:
		if max_score is None or score > max_score:
			max_score = score

	exp_scores = []
	total_sum = 0.0
	for score in scores:
		x = math.exp(score - max_score)
		exp_scores.append(x)
		total_sum += x

	probs = []
	for score in exp_scores:
		probs.append(score / total_sum)
	return probs

def warmup_linear(x, warmup=0.002):
	if x < warmup:
		return x/warmup
	return 1.0 - x


	
def parser_add_argument	( parser ) :
	## Required parameters
	parser.add_argument("--openapi_key", default=None, type=str, required=True, help="OpenAPI key information for morphology analysis")
	parser.add_argument("--bert_model", default=None, type=str, required=True,
						help="Bert pre-trained model selected in the list: bert-base-uncased, "
						"bert-large-uncased, bert-base-cased, bert-large-cased, bert-base-multilingual-uncased, "
						"bert-base-multilingual-cased, bert-base-chinese.")
	
	## Other parameters
	parser.add_argument("--bert_model_file", default=None, type=str, help="specific model file (i.e. pytorch_model.bin)")
	parser.add_argument("--vocab_file", default=None, type=str, help="specific vocab file (i.e. vocab.txt)")
	parser.add_argument("--output_dir", default=None, type=str, 
						help="The output directory where the model checkpoints and predictions will be written.")

	parser.add_argument("--train_file", default=None, type=str, help="SQuAD json for training. E.g., train-v1.1.json")
	parser.add_argument("--predict_file", default=None, type=str,
						help="SQuAD json for predictions. E.g., dev-v1.1.json or test-v1.1.json")
	parser.add_argument("--eval_folder", default=None, type=str, help="SQuAD json folder for test performance")
	parser.add_argument("--max_seq_length", default=384, type=int,
						help="The maximum total input sequence length after WordPiece tokenization. Sequences "
							 "longer than this will be truncated, and sequences shorter than this will be padded.")
	parser.add_argument("--doc_stride", default=128, type=int,
						help="When splitting up a long document into chunks, how much stride to take between chunks.")
	parser.add_argument("--max_query_length", default=64, type=int,
						help="The maximum number of tokens for the question. Questions longer than this will "
							 "be truncated to this length.")
	parser.add_argument("--do_train", action='store_true', help="Whether to run training.")
	parser.add_argument("--do_predict", action='store_true', help="Whether to run eval on the dev set.")
	parser.add_argument("--train_batch_size", default=32, type=int, help="Total batch size for training.")
	parser.add_argument("--predict_batch_size", default=8, type=int, help="Total batch size for predictions.")
	parser.add_argument("--learning_rate", default=5e-5, type=float, help="The initial learning rate for Adam.")
	parser.add_argument("--num_train_epochs", default=3.0, type=float,
						help="Total number of training epochs to perform.")
	parser.add_argument("--warmup_proportion", default=0.1, type=float,
						help="Proportion of training to perform linear learning rate warmup for. E.g., 0.1 = 10% "
							 "of training.")
	parser.add_argument("--n_best_size", default=20, type=int,
						help="The total number of n-best predictions to generate in the nbest_predictions.json "
							 "output file.")
	parser.add_argument("--max_answer_length", default=30, type=int,
						help="The maximum length of an answer that can be generated. This is needed because the start "
							 "and end predictions are not conditioned on one another.")
	parser.add_argument("--verbose_logging", action='store_true',
						help="If true, all of the warnings related to data processing will be printed. "
							 "A number of warnings are expected for a normal SQuAD evaluation.")
	parser.add_argument("--no_cuda",
						action='store_true',
						help="Whether not to use CUDA when available")
	parser.add_argument('--seed',
						type=int,
						default=42,
						help="random seed for initialization")
	parser.add_argument('--gradient_accumulation_steps',
						type=int,
						default=1,
						help="Number of updates steps to accumulate before performing a backward/update pass.")
	parser.add_argument("--do_lower_case",
						action='store_true',
						help="Whether to lower case the input text. True for uncased models, False for cased models.")
	parser.add_argument("--local_rank",
						type=int,
						default=-1,
						help="local_rank for distributed training on gpus")
	parser.add_argument('--fp16',
						action='store_true',
						help="Whether to use 16-bit float precision instead of 32-bit")
	parser.add_argument('--loss_scale',
						type=float, default=0,
						help="Loss scaling to improve fp16 numeric stability. Only used when fp16 set to True.\n"
							 "0 (default value): dynamic loss scaling.\n"
							 "Positive power of 2: static loss scaling value.\n")
	return parser
	
def main():
	parser = argparse.ArgumentParser()
	parser = parser_add_argument( parser )
	args = parser.parse_args()

	if args.local_rank == -1 or args.no_cuda:
		device = torch.device("cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu")
		n_gpu = torch.cuda.device_count()
	else:
		torch.cuda.set_device(args.local_rank)
		device = torch.device("cuda", args.local_rank)
		n_gpu = 1
		# Initializes the distributed backend which will take care of sychronizing nodes/GPUs
		torch.distributed.init_process_group(backend='nccl')
	logger.info("device: {} n_gpu: {}, distributed training: {}, 16-bits training: {}".format(
		device, n_gpu, bool(args.local_rank != -1), args.fp16))

	if args.gradient_accumulation_steps < 1:
		raise ValueError("Invalid gradient_accumulation_steps parameter: {}, should be >= 1".format(
							args.gradient_accumulation_steps))

	args.train_batch_size = int(args.train_batch_size / args.gradient_accumulation_steps)

	random.seed(args.seed)
	np.random.seed(args.seed)
	torch.manual_seed(args.seed)
	if n_gpu > 0:
		torch.cuda.manual_seed_all(args.seed)

	if not args.do_train and not args.do_predict :
		raise ValueError("At least one of `do_train` or `do_predict` must be True.")

	if args.do_train:
		if not args.train_file:
			raise ValueError(
				"If `do_train` is True, then `train_file` must be specified.")
	if args.do_predict:
		if not args.predict_file:
			raise ValueError(
				"If `do_predict` is True, then `predict_file` must be specified.")

	if args.do_train :
		if os.path.exists(args.output_dir) and os.listdir(args.output_dir) :
			logger.error("Output directory () already exists and is not empty.")
		os.makedirs(args.output_dir, exist_ok=True)

	tokenizer_path = args.bert_model
	if args.vocab_file != None :
		tokenizer_path = os.path.join(args.bert_model, args.vocab_file)
	tokenizer = BertTokenizer.from_pretrained(tokenizer_path, do_lower_case=args.do_lower_case)

	train_examples = None
	num_train_steps = None
	if args.do_train:
		#########################################
		### joonho.lim @ 2019-03-15
		train_examples = read_squad_examples_and_do_lang( input_file=args.train_file, openapi_key=args.openapi_key, is_training=True )
		num_train_steps = int(len(train_examples) / args.train_batch_size / args.gradient_accumulation_steps * args.num_train_epochs)
		
		# joonho.lim - for small-size sample training
		if num_train_steps == 0 :
			num_train_steps = 1

	# Prepare model
	state_dict = None
	if args.bert_model_file != None :
		state_dict = torch.load( os.path.join(args.bert_model, args.bert_model_file) )
	model = BertForQuestionAnswering.from_pretrained(args.bert_model, state_dict=state_dict,
				cache_dir=PYTORCH_PRETRAINED_BERT_CACHE / 'distributed_{}'.format(args.local_rank))

	if args.fp16:
		model.half()
	model.to(device)
	if args.local_rank != -1:
		try:
			from apex.parallel import DistributedDataParallel as DDP
		except ImportError:
			raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use distributed and fp16 training.")

		model = DDP(model)
	elif n_gpu > 1:
		model = torch.nn.DataParallel(model)

	# Prepare optimizer
	param_optimizer = list(model.named_parameters())

	# hack to remove pooler, which is not used
	# thus it produce None grad that break apex
	param_optimizer = [n for n in param_optimizer if 'pooler' not in n[0]]

	no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
	optimizer_grouped_parameters = [
		{'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': 0.01},
		{'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
		]

	t_total = num_train_steps
	if args.local_rank != -1:
		t_total = t_total // torch.distributed.get_world_size()
	if args.fp16:
		try:
			from apex.optimizers import FP16_Optimizer
			from apex.optimizers import FusedAdam
		except ImportError:
			raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use distributed and fp16 training.")

		optimizer = FusedAdam(optimizer_grouped_parameters,
							  lr=args.learning_rate,
							  bias_correction=False,
							  max_grad_norm=1.0)
		if args.loss_scale == 0:
			optimizer = FP16_Optimizer(optimizer, dynamic_loss_scale=True)
		else:
			optimizer = FP16_Optimizer(optimizer, static_loss_scale=args.loss_scale)
	else:
		optimizer = BertAdam(optimizer_grouped_parameters,
							 lr=args.learning_rate,
							 warmup=args.warmup_proportion,
							 t_total=t_total)

	global_step = 0
	if args.do_train:
		cached_train_features_file = args.train_file+'_{0}_{1}_{2}_{3}'.format(
			list(filter(None, args.bert_model.split('/'))).pop(), str(args.max_seq_length), str(args.doc_stride), str(args.max_query_length))
		train_features = None
		try:
			with open(cached_train_features_file, "rb") as reader:
				train_features = pickle.load(reader)
		except:
			train_features = convert_examples_to_features(
				examples=train_examples,
				tokenizer=tokenizer,
				max_seq_length=args.max_seq_length,
				doc_stride=args.doc_stride,
				max_query_length=args.max_query_length,
				is_training=True)
			if args.local_rank == -1 or torch.distributed.get_rank() == 0:
				logger.info("  Saving train features into cached file %s", cached_train_features_file)
				with open(cached_train_features_file, "wb") as writer:
					pickle.dump(train_features, writer)
		logger.info("***** Training *****")
		logger.info("  Num orig examples = %d", len(train_examples))
		logger.info("  Num split examples = %d", len(train_features))
		logger.info("  Batch size = %d", args.train_batch_size)
		logger.info("  Num steps = %d", num_train_steps)
		all_input_ids = torch.tensor([f.input_ids for f in train_features], dtype=torch.long)
		all_input_mask = torch.tensor([f.input_mask for f in train_features], dtype=torch.long)
		all_segment_ids = torch.tensor([f.segment_ids for f in train_features], dtype=torch.long)
		all_start_positions = torch.tensor([f.start_position for f in train_features], dtype=torch.long)
		all_end_positions = torch.tensor([f.end_position for f in train_features], dtype=torch.long)
		train_data = TensorDataset(all_input_ids, all_input_mask, all_segment_ids,
								   all_start_positions, all_end_positions)
		if args.local_rank == -1:
			train_sampler = RandomSampler(train_data)
		else:
			train_sampler = DistributedSampler(train_data)
		train_dataloader = DataLoader(train_data, sampler=train_sampler, batch_size=args.train_batch_size)

	model.train()
	for epoch_i in trange(int(args.num_train_epochs), desc="Epoch"):
		for step, batch in enumerate(tqdm(train_dataloader, desc="Iteration")):
			if n_gpu == 1:
				batch = tuple(t.to(device) for t in batch) # multi-gpu does scattering it-self
			input_ids, input_mask, segment_ids, start_positions, end_positions = batch
			loss = model(input_ids, segment_ids, input_mask, start_positions, end_positions)
			if n_gpu > 1:
				loss = loss.mean() # mean() to average on multi-gpu.
			if args.gradient_accumulation_steps > 1:
				loss = loss / args.gradient_accumulation_steps

			if args.fp16:
				optimizer.backward(loss)
			else:
				loss.backward()
			if (step + 1) % args.gradient_accumulation_steps == 0:
				# modify learning rate with special warm up BERT uses
				lr_this_step = args.learning_rate * warmup_linear(global_step/t_total, args.warmup_proportion)
				for param_group in optimizer.param_groups:
					param_group['lr'] = lr_this_step
				optimizer.step()
				optimizer.zero_grad()
				global_step += 1

	# Save a trained model
	model_to_save = model.module if hasattr(model, 'module') else model  # Only save the model it-self
	output_model_file = os.path.join(args.output_dir, "pytorch_model.bin")
	if args.do_train:
		torch.save(model_to_save.state_dict(), output_model_file)

	# Load a trained model that you have fine-tuned
	model_state_dict = torch.load(output_model_file)
	model = BertForQuestionAnswering.from_pretrained(args.bert_model, state_dict=model_state_dict)
	model.to(device)
		
	if args.do_predict and (args.local_rank == -1 or torch.distributed.get_rank() == 0):
		#########################################
		### joonho.lim @ 2019-03-15
		eval_examples = read_squad_examples_and_do_lang( input_file=args.predict_file, openapi_key=args.openapi_key, is_training=False )
		eval_features = convert_examples_to_features(
			examples=eval_examples,
			tokenizer=tokenizer,
			max_seq_length=args.max_seq_length,
			doc_stride=args.doc_stride,
			max_query_length=args.max_query_length,
			is_training=False)

		logger.info("***** Running predictions *****")
		logger.info("  Num orig examples = %d", len(eval_examples))
		logger.info("  Num split examples = %d", len(eval_features))
		logger.info("  Batch size = %d", args.predict_batch_size)

		all_input_ids = torch.tensor([f.input_ids for f in eval_features], dtype=torch.long)
		all_input_mask = torch.tensor([f.input_mask for f in eval_features], dtype=torch.long)
		all_segment_ids = torch.tensor([f.segment_ids for f in eval_features], dtype=torch.long)
		all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)
		eval_data = TensorDataset(all_input_ids, all_input_mask, all_segment_ids, all_example_index)
		# Run prediction for full data
		eval_sampler = SequentialSampler(eval_data)
		eval_dataloader = DataLoader(eval_data, sampler=eval_sampler, batch_size=args.predict_batch_size)

		model.eval()
		all_results = []
		logger.info("Start evaluating")
		for input_ids, input_mask, segment_ids, example_indices in tqdm(eval_dataloader, desc="Evaluating"):
			if len(all_results) % 1000 == 0:
				logger.info("Processing example: %d" % (len(all_results)))
			input_ids = input_ids.to(device)
			input_mask = input_mask.to(device)
			segment_ids = segment_ids.to(device)
			with torch.no_grad():
				batch_start_logits, batch_end_logits = model(input_ids, segment_ids, input_mask)
			for i, example_index in enumerate(example_indices):
				start_logits = batch_start_logits[i].detach().cpu().tolist()
				end_logits = batch_end_logits[i].detach().cpu().tolist()
				eval_feature = eval_features[example_index.item()]
				unique_id = int(eval_feature.unique_id)
				all_results.append(RawResult(unique_id=unique_id,
											 start_logits=start_logits,
											 end_logits=end_logits))
		output_prediction_file = os.path.join(args.output_dir, "predictions.json")
		output_nbest_file = os.path.join(args.output_dir, "nbest_predictions.json")
		write_predictions(eval_examples, eval_features, all_results,
						  args.n_best_size, args.max_answer_length,
						  args.do_lower_case, output_prediction_file,
						  output_nbest_file, args.verbose_logging)
		
		
			
if __name__ == "__main__":
	main()



from pytorch_pretrained_bert import BertTokenizer

import py_ner.bert_ner_train as train
import torch
import py_ner.lstm_cnn_crf_utils as utils

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
trainer = train.BERTNERTrainer()

mode = 'txt'
if mode == 'csv':
    data = "../py_ner/data/ner.csv"
else:
    #for Korean text for now
    data = '../py_ner/data/train.txt'

sentences, tag2idx, labels = trainer.data_processing(data)

#bert-base-multilingual-cased, bert-base-cased
tokenizer_name = 'bert-base-multilingual-cased'
tokenizer = trainer.tokenizer(tokenizer_name)
trainer.data_loading(tokenizer,sentences,tag2idx,labels)
classifier_model_name='bert-base-multilingual-cased'
trainer.load_token_classifier(classifier_model_name,tag2idx)

trainer.set_optimizer()

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
trainer.train_epoch(device)

trainer.eval(device, labels)

language = "kr"
trainer.save_model(language)



# coding=utf-8
from __future__ import print_function

import os
import codecs
import optparse
import itertools
from collections import OrderedDict
import py_ner.ner_data_loader as loader
import torch
import time
import pickle
from torch.autograd import Variable
import matplotlib.pyplot as plt
import sys

import numpy as np

import py_ner.lstm_cnn_crf_utils as utils
import py_ner.ner_data_loader as loader
import py_ner.lstm_cnn_crf_model as model

t = time.time()
models_path = "models/"

optparser = optparse.OptionParser()
optparser.add_option(
    "-T", "--train", default="./eng.train",
    help="Train set location"
)
optparser.add_option(
    "-d", "--dev", default="./eng.testa",
    help="Dev set location"
)
optparser.add_option(
    "-t", "--test", default="./eng.testb",
    help="Test set location"
)
optparser.add_option(
    '--test_train', default='./eng.train54019',
    help='test train'
)
optparser.add_option(
    '--score', default='./temp/score.txt',
    help='score file location'
)
optparser.add_option(
    "-s", "--tag_scheme", default="iobes",
    help="Tagging scheme (IOB or IOBES)"
)
optparser.add_option(
    "-l", "--lower", default="1",
    type='int', help="Lowercase words (this will not affect character inputs)"
)
optparser.add_option(
    "-z", "--zeros", default="0",
    type='int', help="Replace digits with 0"
)
optparser.add_option(
    "-c", "--char_dim", default="25",
    type='int', help="Char embedding dimension"
)
optparser.add_option(
    "-C", "--char_lstm_dim", default="25",
    type='int', help="Char LSTM hidden layer size"
)
optparser.add_option(
    "-b", "--char_bidirect", default="1",
    type='int', help="Use a bidirectional LSTM for chars"
)
optparser.add_option(
    "-w", "--word_dim", default="100",
    type='int', help="Token embedding dimension"
)
optparser.add_option(
    "-W", "--word_lstm_dim", default="200",
    type='int', help="Token LSTM hidden layer size"
)
optparser.add_option(
    "-B", "--word_bidirect", default="1",
    type='int', help="Use a bidirectional LSTM for words"
)
optparser.add_option(
    "-p", "--pre_emb", default="../embeddings/glove.6B.100d.txt",
    help="Location of pretrained embeddings"
)
optparser.add_option(
    "-A", "--all_emb", default="1",
    type='int', help="Load all embeddings"
)
optparser.add_option(
    "-a", "--cap_dim", default="0",
    type='int', help="Capitalization feature dimension (0 to disable)"
)
optparser.add_option(
    "-f", "--crf", default="1",
    type='int', help="Use CRF (0 to disable)"
)
optparser.add_option(
    "-D", "--dropout", default="0.5",
    type='float', help="Droupout on the input (0 = no dropout)"
)
optparser.add_option(
    "-r", "--reload", default="0",
    type='int', help="Reload the last saved model"
)
optparser.add_option(
    "-g", '--use_gpu', default='0',
    type='int', help='whether or not to ues gpu'
)
optparser.add_option(
    '--loss', default='loss.txt',
    help='loss file location'
)
optparser.add_option(
    '--name', default='lstm_cnn_crf',
    help='model name'
)
optparser.add_option(
    '--char_mode', choices=['CNN', 'LSTM'], default='CNN',
    help='char_CNN or char_LSTM'
)
opts = optparser.parse_args()[0]

parameters = OrderedDict()
parameters['tag_scheme'] = opts.tag_scheme
parameters['lower'] = opts.lower == 1
parameters['zeros'] = opts.zeros == 1
parameters['char_dim'] = opts.char_dim
parameters['char_lstm_dim'] = opts.char_lstm_dim
parameters['char_bidirect'] = opts.char_bidirect == 1
parameters['word_dim'] = opts.word_dim
parameters['word_lstm_dim'] = opts.word_lstm_dim
parameters['word_bidirect'] = opts.word_bidirect == 1
parameters['pre_emb'] = opts.pre_emb
parameters['all_emb'] = opts.all_emb == 1
parameters['cap_dim'] = opts.cap_dim
parameters['crf'] = opts.crf == 1
parameters['dropout'] = opts.dropout
parameters['reload'] = opts.reload == 1
parameters['name'] = opts.name
parameters['char_mode'] = opts.char_mode

parameters['use_gpu'] = opts.use_gpu == 1 and torch.cuda.is_available()
use_gpu = parameters['use_gpu']

mapping_file = 'models/mapping.pkl'

name = parameters['name']
model_name = models_path + name #get_name(parameters)
tmp_model = model_name + '.tmp'


assert os.path.isfile(opts.train)
assert os.path.isfile(opts.dev)
assert os.path.isfile(opts.test)
assert parameters['char_dim'] > 0 or parameters['word_dim'] > 0
assert 0. <= parameters['dropout'] < 1.0
assert parameters['tag_scheme'] in ['iob', 'iobes']
assert not parameters['all_emb'] or parameters['pre_emb']
assert not parameters['pre_emb'] or parameters['word_dim'] > 0
#assert not parameters['pre_emb'] or os.path.isfile(parameters['pre_emb'])

if not os.path.isfile(utils.eval_script):
    raise Exception('CoNLL evaluation script not found at "%s"' % utils.eval_script)
if not os.path.exists(model.eval_temp):
    os.makedirs(utils.eval_temp)
if not os.path.exists(models_path):
    os.makedirs(models_path)

lower = parameters['lower']
zeros = parameters['zeros']
tag_scheme = parameters['tag_scheme']

train_sentences = loader.load_sentences(opts.train, lower, zeros)
dev_sentences = loader.load_sentences(opts.dev, lower, zeros)
test_sentences = loader.load_sentences(opts.test, lower, zeros)
test_train_sentences = loader.load_sentences(opts.test_train, lower, zeros)

loader.update_tag_scheme(train_sentences, tag_scheme)
loader.update_tag_scheme(dev_sentences, tag_scheme)
loader.update_tag_scheme(test_sentences, tag_scheme)
loader.update_tag_scheme(test_train_sentences, tag_scheme)

dico_words_train = loader.word_mapping(train_sentences, lower)[0]

dico_words, word_to_id, id_to_word = loader.augment_with_pretrained(
        dico_words_train.copy(),
        parameters['pre_emb'],
        list(itertools.chain.from_iterable(
            [[w[0] for w in s] for s in dev_sentences + test_sentences])
        ) if not parameters['all_emb'] else None
    )

dico_chars, char_to_id, id_to_char = loader.char_mapping(train_sentences)
dico_tags, tag_to_id, id_to_tag = loader.tag_mapping(train_sentences)

train_data = loader.prepare_dataset(
    train_sentences, word_to_id, char_to_id, tag_to_id, lower
)
dev_data = loader.prepare_dataset(
    dev_sentences, word_to_id, char_to_id, tag_to_id, lower
)
test_data = loader.prepare_dataset(
    test_sentences, word_to_id, char_to_id, tag_to_id, lower
)
test_train_data = loader.prepare_dataset(
    test_train_sentences, word_to_id, char_to_id, tag_to_id, lower
)

print("%i / %i / %i sentences in train / dev / test." % (
    len(train_data), len(dev_data), len(test_data)))

all_word_embeds = {}
for i, line in enumerate(codecs.open(opts.pre_emb, 'r', 'utf-8')):
    s = line.strip().split()
    if len(s) == parameters['word_dim'] + 1:
        all_word_embeds[s[0]] = np.array([float(i) for i in s[1:]])

word_embeds = np.random.uniform(-np.sqrt(0.06), np.sqrt(0.06), (len(word_to_id), opts.word_dim))

for w in word_to_id:
    if w in all_word_embeds:
        word_embeds[word_to_id[w]] = all_word_embeds[w]
    elif w.lower() in all_word_embeds:
        word_embeds[word_to_id[w]] = all_word_embeds[w.lower()]

print('Loaded %i pretrained embeddings.' % len(all_word_embeds))

with open(mapping_file, 'wb') as f:
    mappings = {
        'word_to_id': word_to_id,
        'tag_to_id': tag_to_id,
        'char_to_id': char_to_id,
        'parameters': parameters,
        'word_embeds': word_embeds
    }
    pickle.dump(mappings, f)

print('word_to_id: ', len(word_to_id))
model = model.BiLSTM_CRF(vocab_size=len(word_to_id),
                   tag_to_ix=tag_to_id,
                   embedding_dim=parameters['word_dim'],
                   hidden_dim=parameters['word_lstm_dim'],
                   use_gpu=use_gpu,
                   char_to_ix=char_to_id,
                   pre_word_embeds=word_embeds,
                   use_crf=parameters['crf'],
                   char_mode=parameters['char_mode'])
                   # n_cap=4,
                   # cap_embedding_dim=10)
if parameters['reload']:
    model.load_state_dict(torch.load(model_name))

if use_gpu:
    model.cuda()

learning_rate = 0.015

optimizer = torch.optim.SGD(model.parameters(), lr=learning_rate, momentum=0.9)
losses = []
loss = 0.0
best_dev_F = -1.0
best_test_F = -1.0
best_train_F = -1.0
all_F = [[0, 0, 0]]
plot_every = 10
eval_every = 20
count = 0

def evaluating(model, datas, best_F):
    prediction = []
    save = False
    new_F = 0.0
    confusion_matrix = torch.zeros((len(tag_to_id) - 2, len(tag_to_id) - 2))
    for data in datas:
        ground_truth_id = data['tags']
        words = data['str_words']
        chars2 = data['chars']
        caps = data['caps']

        if parameters['char_mode'] == 'LSTM':
            chars2_sorted = sorted(chars2, key=lambda p: len(p), reverse=True)
            d = {}
            for i, ci in enumerate(chars2):
                for j, cj in enumerate(chars2_sorted):
                    if ci == cj and not j in d and not i in d.values():
                        d[j] = i
                        continue
            chars2_length = [len(c) for c in chars2_sorted]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_sorted), char_maxl), dtype='int')
            for i, c in enumerate(chars2_sorted):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))

        if parameters['char_mode'] == 'CNN':
            d = {}
            chars2_length = [len(c) for c in chars2]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_length), char_maxl), dtype='int')
            for i, c in enumerate(chars2):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))

        dwords = Variable(torch.LongTensor(data['words']))
        dcaps = Variable(torch.LongTensor(caps))
        if use_gpu:
            val, out = model(dwords.cuda(), chars2_mask.cuda(), dcaps.cuda(), chars2_length, d)
        else:
            val, out = model(dwords, chars2_mask, dcaps, chars2_length, d)

        predicted_id = out
        for (word, true_id, pred_id) in zip(words, ground_truth_id, predicted_id):
            line = ' '.join([word, id_to_tag[true_id], id_to_tag[pred_id]])
            prediction.append(line)
            confusion_matrix[true_id, pred_id] += 1
        prediction.append('')

    predf = utils.eval_temp + '/pred.' + name
    scoref = utils.eval_temp + '/score.' + name

    with open(predf, 'w') as f:
        f.write('\n'.join(prediction))

    os.system('%s < %s > %s' % (utils.eval_script, predf, scoref))
    eval_lines = [l.rstrip() for l in codecs.open(scoref, 'r', 'utf8')]

    for i, line in enumerate(eval_lines):
        if i == 1:
            new_F = float(line.strip().split()[-1])
            if new_F > best_F:
                best_F = new_F
                save = True
                print('the best F is ', new_F)

    print(("{: >2}{: >7}{: >7}%s{: >9}" % ("{: >7}" * confusion_matrix.size(0))).format(
        "ID", "NE", "Total", *([id_to_tag[i] for i in range(confusion_matrix.size(0))] + ["Percent"])
    ))
    for i in range(confusion_matrix.size(0)):
        print(("{: >2}{: >7}{: >7}%s{: >9}" % ("{: >7}" * confusion_matrix.size(0))).format(
            str(i), id_to_tag[i], str(confusion_matrix[i].sum()),
            *([confusion_matrix[i][j] for j in range(confusion_matrix.size(0))] +
              ["%.3f" % (confusion_matrix[i][i] * 100. / max(1, confusion_matrix[i].sum()))])
        ))
    return best_F, new_F, save

saved = False
model.train(True)
#for epoch in range(1, 10001):
for epoch in range(1, 2):
    for i, index in enumerate(np.random.permutation(len(train_data))):
        tr = time.time()
        count += 1
        data = train_data[index]
        model.zero_grad()

        sentence_in = data['words']
        sentence_in = Variable(torch.LongTensor(sentence_in))
        tags = data['tags']
        chars2 = data['chars']

        ######### char lstm
        if parameters['char_mode'] == 'LSTM':
            chars2_sorted = sorted(chars2, key=lambda p: len(p), reverse=True)
            d = {}
            for i, ci in enumerate(chars2):
                for j, cj in enumerate(chars2_sorted):
                    if ci == cj and not j in d and not i in d.values():
                        d[j] = i
                        continue
            chars2_length = [len(c) for c in chars2_sorted]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_sorted), char_maxl), dtype='int')
            for i, c in enumerate(chars2_sorted):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))

        # ######## char cnn
        if parameters['char_mode'] == 'CNN':
            d = {}
            chars2_length = [len(c) for c in chars2]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_length), char_maxl), dtype='int')
            for i, c in enumerate(chars2):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))


        targets = torch.LongTensor(tags)
        caps = Variable(torch.LongTensor(data['caps']))
        if use_gpu:
            neg_log_likelihood = model.neg_log_likelihood(sentence_in.cuda(), targets.cuda(), chars2_mask.cuda(), caps.cuda(), chars2_length, d)
        else:
            neg_log_likelihood = model.neg_log_likelihood(sentence_in, targets, chars2_mask, caps, chars2_length, d)
        loss += neg_log_likelihood.data / len(data['words'])

        neg_log_likelihood.backward()

        torch.nn.utils.clip_grad_norm(model.parameters(), 5.0)
        optimizer.step()

        if count % plot_every == 0:
            loss /= plot_every
            print(count, ': ', loss)
            if losses == []:
                losses.append(loss)
            losses.append(loss)
            text = '<p>' + '</p><p>'.join([str(l) for l in losses[-9:]]) + '</p>'
            losswin = 'loss_' + name
            textwin = 'loss_text_' + name

            loss = 0.0

        if count % (eval_every) == 0 and count > (eval_every * 20) or \
                count % (eval_every*4) == 0 and count < (eval_every * 20):
            model.train(False)
            best_train_F, new_train_F, _ = evaluating(model, test_train_data, best_train_F)
            best_dev_F, new_dev_F, save = evaluating(model, dev_data, best_dev_F)
            if save:
                saved = True
                print('saving the model....')
                torch.save(model, model_name)

            best_test_F, new_test_F, _ = evaluating(model, test_data, best_test_F)
            all_F.append([new_train_F, new_dev_F, new_test_F])
            Fwin = 'F-score of {train, dev, test}_' + name

            model.train(True)

        if count % len(train_data) == 0:
            utils.adjust_learning_rate(optimizer, lr=learning_rate / (1 + 0.05 * count / len(train_data)))

if saved == False:
    print('saving the model....')
    torch.save(model, model_name)

print(time.time() - t)

plt.plot(losses)
plt.show()







import numpy as np # linear algebra
import pandas as pd # data processing, CSV file I/O
from seqeval.metrics import f1_score
import os
from py_ner.ner_data_loader import SentenceGetter, convert_to_df, read_file, convert_to_df_for_ko_ner
import torch
from torch.optim import Adam
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, SequentialSampler
from keras.preprocessing.sequence import pad_sequences
from sklearn.model_selection import train_test_split
from pytorch_pretrained_bert import BertTokenizer, BertConfig
from pytorch_pretrained_bert import BertForTokenClassification, BertAdam
from tqdm import tqdm, trange
import py_ner.lstm_cnn_crf_utils as utils
import pickle

class BERTNERTrainer:
    def __init__(self):
        print('BertNERTrainer')
        self.model = None
        self.train_dataloader = None
        self.valid_dataloader = None
        self.optimizer = None

    def data_processing(self, data, mode='txt'):
        #mode = 'kr' # kr or en
        if mode == 'csv':
            dataset = convert_to_df(data)
            json_file = './models/tag_info_en.json'
        elif mode == 'txt':
            print("here!!!!!!")
            #data = './data/train.txt'
            json_file = './models/tag_info_kr.json'
            train_sents = read_file(data)
            dataset, tag_names = convert_to_df_for_ko_ner(train_sents)

            print(dataset.info())
            getter = SentenceGetter(dataset)
            sentences = [" ".join([s[0] for s in sent]) for sent in getter.sentences]
            print(sentences[0])

            labels = [[s[2] for s in sent] for sent in getter.sentences]

            tags_vals = list(set(dataset["tag"].values))
            tag2idx = {t: i for i, t in enumerate(tags_vals)}
            print(str(sentences[0]) + ' --> ' + str(labels[0]))
            print(str(tags_vals) + " :: " + str(tag2idx))

            utils.save_to_json(tag2idx, json_file)

        return sentences, tag2idx, labels

    def tokenizer(self, tokenizer_name):
        tokenizer = BertTokenizer.from_pretrained(tokenizer_name, do_lower_case=True)
        return tokenizer

    def data_loading(self, tokenizer, sentences, tag2idx, labels):
        MAX_LEN = 40
        bs =64

        tokenized_texts = [tokenizer.tokenize(sent) for sent in sentences]
        print(tokenized_texts[0])

        input_ids = pad_sequences([tokenizer.convert_tokens_to_ids(txt) for txt in tokenized_texts],
                                  maxlen=MAX_LEN, dtype="long", truncating="post", padding="post")

        tags = pad_sequences([[tag2idx.get(l) for l in lab] for lab in labels],
                             maxlen=MAX_LEN, value=tag2idx["O"], padding="post",
                             dtype="long", truncating="post")

        attention_masks = [[float(i>0) for i in ii] for ii in input_ids]

        tr_inputs, val_inputs, tr_tags, val_tags = train_test_split(input_ids, tags,
                                                                    random_state=2018, test_size=0.1)
        tr_masks, val_masks, _, _ = train_test_split(attention_masks, input_ids,
                                                     random_state=2018, test_size=0.1)

        tr_inputs = torch.tensor(tr_inputs)
        val_inputs = torch.tensor(val_inputs)
        tr_tags = torch.tensor(tr_tags)
        val_tags = torch.tensor(val_tags)
        tr_masks = torch.tensor(tr_masks)
        val_masks = torch.tensor(val_masks)

        train_data = TensorDataset(tr_inputs, tr_masks, tr_tags)
        train_sampler = RandomSampler(train_data)
        self.train_dataloader = DataLoader(train_data, sampler=train_sampler, batch_size=bs)

        valid_data = TensorDataset(val_inputs, val_masks, val_tags)
        valid_sampler = SequentialSampler(valid_data)
        self.valid_dataloader = DataLoader(valid_data, sampler=valid_sampler, batch_size=bs)

    def load_token_classifier(self, classifier_model_name, tag2idx):
        self.model = BertForTokenClassification.from_pretrained(classifier_model_name, num_labels=len(tag2idx))

    def set_optimizer(self):
        FULL_FINETUNING = True
        if FULL_FINETUNING:
            param_optimizer = list(self.model.named_parameters())
            no_decay = ['bias', 'gamma', 'beta']
            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
                 'weight_decay_rate': 0.01},
                {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
                 'weight_decay_rate': 0.0}
            ]
        else:
            param_optimizer = list(self.model.classifier.named_parameters())
            optimizer_grouped_parameters = [{"params": [p for n, p in param_optimizer]}]

        self.optimizer = Adam(optimizer_grouped_parameters, lr=3e-5)

    def flat_accuracy(self, preds, labels):
        pred_flat = np.argmax(preds, axis=2).flatten()
        labels_flat = labels.flatten()
        return np.sum(pred_flat == labels_flat) / len(labels_flat)

    def train_epoch(self, device):
        epochs = 5
        max_grad_norm = 1.0

        for _ in trange(epochs, desc="Epoch"):
            # TRAIN loop
            self.model.train()
            tr_loss = 0
            nb_tr_examples, nb_tr_steps = 0, 0
            for step, batch in enumerate(self.train_dataloader):
                # add batch to gpu
                batch = tuple(t.to(device) for t in batch)
                b_input_ids, b_input_mask, b_labels = batch
                b_input_ids = torch.tensor(b_input_ids).to(device).long()
                b_labels = torch.tensor(b_labels).to(device).long()

                # forward pass
                loss = self.model(b_input_ids, token_type_ids=None,
                             attention_mask=b_input_mask, labels=b_labels)
                # backward pass
                loss.backward()
                # track train loss
                tr_loss += loss.item()
                nb_tr_examples += b_input_ids.size(0)
                nb_tr_steps += 1
                # gradient clipping
                torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=max_grad_norm)
                # update parameters
                self.optimizer.step()
                self.model.zero_grad()
                # print train loss per epoch
            print("Train loss: {}".format(tr_loss / nb_tr_steps))

    def eval(self, device, tags_vals):
        # VALIDATION on validation set
        self.model.eval()
        eval_loss, eval_accuracy = 0, 0
        nb_eval_steps, nb_eval_examples = 0, 0
        predictions, true_labels = [], []
        for batch in self.valid_dataloader:
            batch = tuple(t.to(device) for t in batch)
            b_input_ids, b_input_mask, b_labels = batch
            b_input_ids = torch.tensor(b_input_ids).to(device).long()
            b_labels = torch.tensor(b_labels).to(device).long()

            with torch.no_grad():
                tmp_eval_loss = self.model(b_input_ids, token_type_ids=None,
                                      attention_mask=b_input_mask, labels=b_labels)
                logits = self.model(b_input_ids, token_type_ids=None,
                               attention_mask=b_input_mask)
            logits = logits.detach().cpu().numpy()
            label_ids = b_labels.to('cpu').numpy()
            predictions.extend([list(p) for p in np.argmax(logits, axis=2)])
            true_labels.append(label_ids)

            tmp_eval_accuracy = self.flat_accuracy(logits, label_ids)

            eval_loss += tmp_eval_loss.mean().item()
            eval_accuracy += tmp_eval_accuracy

            nb_eval_examples += b_input_ids.size(0)
            nb_eval_steps += 1
        eval_loss = eval_loss / nb_eval_steps
        print("Validation loss: {}".format(eval_loss))
        print("Validation Accuracy: {}".format(eval_accuracy / nb_eval_steps))
        pred_tags = [tags_vals[p_i] for p in predictions for p_i in p]
        valid_tags = [tags_vals[l_ii] for l in true_labels for l_i in l for l_ii in l_i]
        print("F1-Score: {}".format(f1_score(pred_tags, valid_tags)))

    def save_model(self, language):
        if language == 'kr' or language == 'ko':
            pickle.dump(self.model, open('bert_ner_kr.model', 'wb'))
        else:
            pickle.dump(self.model, open('bert_ner_en.model', 'wb'))






import pyTextMiner as ptm

pipeline = None
corpus = ptm.CorpusFromFieldDelimitedFile('./data/donald.txt',2)
mecab_path = 'C:\\mecab\\mecab-ko-dic'
mode = 'korean_lemmatizer'
if mode is not 'korean_lemmatizer':
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.MeCab(mecab_path),
                            #ptm.tokenizer.Komoran(),
                            ptm.helper.SelectWordOnly(),
                            ptm.ngram.NGramTokenizer(1,2,concat=' '),
                            ptm.helper.StopwordFilter(file='./stopwords/stopwordsKor.txt'))
else :
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.MeCab(mecab_path),
                            #ptm.tokenizer.Komoran(),
                            ptm.lemmatizer.SejongPOSLemmatizer(),
                            ptm.helper.SelectWordOnly(),
                            # ptm.ngram.NGramTokenizer(1, 2, concat=' '))
                            ptm.helper.StopwordFilter(file='./stopwords/stopwordsKor.txt'))


documents = ['오늘은 비가와서 그런지 매우 우울하다',
             '시험이 끝나야 놀지 스트레스 받아ㅠㅠ',
             '행복한 하루의 끝이라 아름답고 좋네!',
             '더운날에는 아이스 커피가 최고지~~!']

#result = pipeline.processCorpus(corpus)
result = pipeline.processCorpus(documents)
print(result)


# coding=utf-8
from __future__ import print_function
import optparse
import os
import pickle

import torch
import time
import numpy as np

from torch.autograd import Variable

import py_ner.ner_data_loader as loader
import  py_ner.lstm_cnn_crf_utils as utils

t = time.time()

# python -m visdom.server

optparser = optparse.OptionParser()
optparser.add_option(
    "-t", "--test", default="./eng.testb",
    help="Test set location"
)
optparser.add_option(
    '--score', default='./temp/score.txt',
    help='score file location'
)
optparser.add_option(
    "-f", "--crf", default="0",
    type='int', help="Use CRF (0 to disable)"
)
optparser.add_option(
    "-g", '--use_gpu', default='1',
    type='int', help='whether or not to ues gpu'
)
optparser.add_option(
    '--loss', default='loss.txt',
    help='loss file location'
)
optparser.add_option(
    '--model_path', default='models/lstm_cnn_crf',
    help='model path'
)
optparser.add_option(
    '--map_path', default='models/mapping.pkl',
    help='model path'
)
optparser.add_option(
    '--char_mode', choices=['CNN', 'LSTM'], default='CNN',
    help='char_CNN or char_LSTM'
)

opts = optparser.parse_args()[0]

mapping_file = opts.map_path

with open(mapping_file, 'rb') as f:
    mappings = pickle.load(f)

word_to_id = mappings['word_to_id']
tag_to_id = mappings['tag_to_id']
id_to_tag = {k[1]: k[0] for k in tag_to_id.items()}
char_to_id = mappings['char_to_id']
parameters = mappings['parameters']
word_embeds = mappings['word_embeds']

use_gpu =  opts.use_gpu == 1 and torch.cuda.is_available()


assert os.path.isfile(opts.test)
assert parameters['tag_scheme'] in ['iob', 'iobes']

if not os.path.isfile(utils.eval_script):
    raise Exception('CoNLL evaluation script not found at "%s"' % utils.eval_script)
if not os.path.exists(utils.eval_temp):
    os.makedirs(utils.eval_temp)

lower = parameters['lower']
zeros = parameters['zeros']
tag_scheme = parameters['tag_scheme']

test_sentences = loader.load_sentences(opts.test, lower, zeros)
loader.update_tag_scheme(test_sentences, tag_scheme)
test_data = loader.prepare_dataset(
    test_sentences, word_to_id, char_to_id, tag_to_id, lower
)

model = torch.load(opts.model_path)
model_name = opts.model_path.split('/')[-1].split('.')[0]

if use_gpu:
    model.cuda()

model.eval()

# def getmaxlen(tags):
#     l = 1
#     maxl = 1
#     for tag in tags:
#         tag = id_to_tag[tag]
#         if 'I-' in tag:
#             l += 1
#         elif 'B-' in tag:
#             l = 1
#         elif 'E-' in tag:
#             l += 1
#             maxl = max(maxl, l)
#             l = 1
#         elif 'O' in tag:
#             l = 1
#     return maxl

def eval(model, datas, maxl=1):
    prediction = []
    confusion_matrix = torch.zeros((len(tag_to_id) - 2, len(tag_to_id) - 2))
    for data in datas:
        ground_truth_id = data['tags']
        # l = getmaxlen(ground_truth_id)
        # if not l == maxl:
        #     continue
        words = data['str_words']
        chars2 = data['chars']
        caps = data['caps']

        if parameters['char_mode'] == 'LSTM':
            chars2_sorted = sorted(chars2, key=lambda p: len(p), reverse=True)
            d = {}
            for i, ci in enumerate(chars2):
                for j, cj in enumerate(chars2_sorted):
                    if ci == cj and not j in d and not i in d.values():
                        d[j] = i
                        continue
            chars2_length = [len(c) for c in chars2_sorted]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_sorted), char_maxl), dtype='int')
            for i, c in enumerate(chars2_sorted):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))

        if parameters['char_mode'] == 'CNN':
            d = {}
            chars2_length = [len(c) for c in chars2]
            char_maxl = max(chars2_length)
            chars2_mask = np.zeros((len(chars2_length), char_maxl), dtype='int')
            for i, c in enumerate(chars2):
                chars2_mask[i, :chars2_length[i]] = c
            chars2_mask = Variable(torch.LongTensor(chars2_mask))

        dwords = Variable(torch.LongTensor(data['words']))
        dcaps = Variable(torch.LongTensor(caps))
        if use_gpu:
            val, out = model(dwords.cuda(), chars2_mask.cuda(), dcaps.cuda(),chars2_length, d)
        else:
            val, out = model(dwords, chars2_mask, dcaps, chars2_length, d)
        predicted_id = out
        for (word, true_id, pred_id) in zip(words, ground_truth_id, predicted_id):
            line = ' '.join([word, id_to_tag[true_id], id_to_tag[pred_id]])
            prediction.append(line)
            confusion_matrix[true_id, pred_id] += 1
        prediction.append('')
    predf = utils.eval_temp + '/pred.' + model_name
    scoref = utils.eval_temp + '/score.' + model_name

    with open(predf, 'w') as f:
        f.write('\n'.join(prediction))

    os.system('%s < %s > %s' % (utils.eval_script, predf, scoref))

    with open(scoref, 'rb') as f:
        for l in f.readlines():
            print(l.strip())

    print(("{: >2}{: >7}{: >7}%s{: >9}" % ("{: >7}" * confusion_matrix.size(0))).format(
        "ID", "NE", "Total",
        *([id_to_tag[i] for i in range(confusion_matrix.size(0))] + ["Percent"])
    ))
    for i in range(confusion_matrix.size(0)):
        print(("{: >2}{: >7}{: >7}%s{: >9}" % ("{: >7}" * confusion_matrix.size(0))).format(
            str(i), id_to_tag[i], str(confusion_matrix[i].sum()),
            *([confusion_matrix[i][j] for j in range(confusion_matrix.size(0))] +
              ["%.3f" % (confusion_matrix[i][i] * 100. / max(1, confusion_matrix[i].sum()))])
        ))

eval(model, test_data, 200)
# for l in range(1, 6):
#     print('maxl=', l)
#     eval(model, test_data, l)
#     # print()
# # for i in range(10):
# #     eval(model, test_data, 100)

print(time.time() - t)


'''
model_testing_sentences = ['Jay is from India', 'Donald is the president of USA']

# parameters
lower = parameters['lower']

# preprocessing
final_test_data = []
for sentence in model_testing_sentences:
    s = sentence.split()
    str_words = [w for w in s]
    words = [word_to_id[utils.lower_case(w, lower) if utils.lower_case(w, lower) in word_to_id else '<UNK>'] for w in str_words]

    # Skip characters that are not in the training set
    chars = [[char_to_id[c] for c in w if c in char_to_id] for w in str_words]

    final_test_data.append({
        'str_words': str_words,
        'words': words,
        'chars': chars,
    })

# prediction
predictions = []
print("Prediction:")
print("word : tag")
for data in final_test_data:
    words = data['str_words']
    chars2 = data['chars']

    d = {}

    # Padding the each word to max word size of that sentence
    chars2_length = [len(c) for c in chars2]
    char_maxl = max(chars2_length)
    chars2_mask = np.zeros((len(chars2_length), char_maxl), dtype='int')
    for i, c in enumerate(chars2):
        chars2_mask[i, :chars2_length[i]] = c
    chars2_mask = Variable(torch.LongTensor(chars2_mask))

    dwords = Variable(torch.LongTensor(data['words']))

    # We are getting the predicted output from our model
    if use_gpu:
        val, predicted_id = model(dwords.cuda(), chars2_mask.cuda(), chars2_length, d)
    else:
        val, predicted_id = model(dwords, chars2_mask, chars2_length, d)

    pred_chunks = utils.get_chunks(predicted_id, tag_to_id)
    temp_list_tags = ['NA'] * len(words)
    for p in pred_chunks:
        temp_list_tags[p[1]] = p[0]

    for word, tag in zip(words, temp_list_tags):
        print(word, ':', tag)
    print('\n')
'''


"""Module for postprocessing and displaying transformer attentions.

"""

import json
from py_ner.bertviz.attention import get_attention

import os

def show(model, model_type, tokenizer, decoding_ner_sentence, sentence_a, sentence_b=None):

    if sentence_b:
        vis_html = """
          <span style="user-select:none">
            Layer: <select id="layer"></select>
            Attention: <select id="filter">
              <option value="all">All</option>
              <option value="aa">Sentence A -> Sentence A</option>
              <option value="ab">Sentence A -> Sentence B</option>
              <option value="ba">Sentence B -> Sentence A</option>
              <option value="bb">Sentence B -> Sentence B</option>
            </select>
          </span>
          <div id='vis'></div> 
        """
    else:
        vis_html = """
          <span style="user-select:none">
            Layer: <select id="layer"></select>
            Attention: <select id="filter">
              <option value="all">All</option>
              <option value="aa">Sentence A -> Sentence A</option>
            </select>
          </span>
          <div id='vis'></div> 
        """

    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__)))
    vis_js = open(os.path.join(__location__, 'head_view.js')).read()
    attn_data = get_attention(model, model_type, tokenizer, sentence_a, sentence_b)
    params = {
        'attention': attn_data,
        'default_filter': "all"
    }

    with open('bert_visualization.html', 'w') as f:
        _head = "<html><meta charset=\"utf-8\"><head><title>BERT Attention Visualization</title>" \
                "<script src=\"https://requirejs.org/docs/release/2.3.6/minified/require.js\" type=\"text/javascript\"></script></head>\n<body>"

        _body = "<script>\n" \
                "require.config({\n" \
                    "paths: {\n" \
                                "\"d3\": \"https//cdnjs.cloudflare.com/ajax/libs/d3/3.4.8/d3.min\",\n" \
                                "\"jquery\": \"https//ajax.googleapis.com/ajax/libs/jquery/2.0.0/jquery.min\"," \
                            "}\n" \
                        "});\n"

        f.write(_head)

        f.write('\n' + vis_html + '\n')
        f.write('\n' + "<h2>")
        f.write(decoding_ner_sentence)
        f.write('\n' + "</h2>\n")

        f.write(_body)

        f.write('window.params = %s' % json.dumps(params) + '\n')
        #f.write("\n</script>")
        #f.write("<script>")
        f.write(vis_js + '\n')
        f.write("\n</script>")

        f.write("</body></html>")



#Loading the data set - training data.
from sklearn.datasets import fetch_20newsgroups
twenty_train = fetch_20newsgroups(subset='train', shuffle=True)


# In[4]:

# You can check the target names (categories) and some data files by following commands.
twenty_train.target_names #prints all the categories


print("\n".join(twenty_train.data[0].split("\n")[:3])) #prints first line of the first data file

# Extracting features from text files
from sklearn.feature_extraction.text import CountVectorizer

count_vect = CountVectorizer()
X_train_counts = count_vect.fit_transform(twenty_train.data)
X_train_counts.shape

# In[7]:

# TF-IDF
from sklearn.feature_extraction.text import TfidfTransformer

tfidf_transformer = TfidfTransformer()
X_train_tfidf = tfidf_transformer.fit_transform(X_train_counts)
X_train_tfidf.shape

# In[9]:

# Machine Learning
# Training Naive Bayes (NB) classifier on training data.
from sklearn.naive_bayes import MultinomialNB

clf = MultinomialNB().fit(X_train_tfidf, twenty_train.target)

# In[14]:

# Building a pipeline: We can write less code and do all of the above, by building a pipeline as follows:
# The names ‘vect’ , ‘tfidf’ and ‘clf’ are arbitrary but will be used later.
# We will be using the 'text_clf' going forward.
from sklearn.pipeline import Pipeline

text_clf = Pipeline([('vect', CountVectorizer()), ('tfidf', TfidfTransformer()), ('clf', MultinomialNB())])

text_clf = text_clf.fit(twenty_train.data, twenty_train.target)

# In[15]:

# Performance of NB Classifier
import numpy as np

twenty_test = fetch_20newsgroups(subset='test', shuffle=True)
predicted = text_clf.predict(twenty_test.data)
np.mean(predicted == twenty_test.target)

# In[16]:

# Training Support Vector Machines - SVM and calculating its performance

from sklearn.linear_model import SGDClassifier

text_clf_svm = Pipeline([('vect', CountVectorizer()), ('tfidf', TfidfTransformer()),
                         ('clf-svm', SGDClassifier(loss='hinge', penalty='l2', alpha=1e-3, n_iter=5, random_state=42))])

text_clf_svm = text_clf_svm.fit(twenty_train.data, twenty_train.target)
predicted_svm = text_clf_svm.predict(twenty_test.data)
np.mean(predicted_svm == twenty_test.target)

# In[18]:

# Grid Search
# Here, we are creating a list of parameters for which we would like to do performance tuning.
# All the parameters name start with the classifier name (remember the arbitrary name we gave).
# E.g. vect__ngram_range; here we are telling to use unigram and bigrams and choose the one which is optimal.

from sklearn.model_selection import GridSearchCV

parameters = {'vect__ngram_range': [(1, 1), (1, 2)], 'tfidf__use_idf': (True, False), 'clf__alpha': (1e-2, 1e-3)}

# In[19]:

# Next, we create an instance of the grid search by passing the classifier, parameters
# and n_jobs=-1 which tells to use multiple cores from user machine.

gs_clf = GridSearchCV(text_clf, parameters, n_jobs=-0)
gs_clf = gs_clf.fit(twenty_train.data, twenty_train.target)

# In[23]:

# To see the best mean score and the params, run the following code

gs_clf.best_score_
gs_clf.best_params_

# Output for above should be: The accuracy has now increased to ~90.6% for the NB classifier (not so naive anymore! 😄)
# and the corresponding parameters are {‘clf__alpha’: 0.01, ‘tfidf__use_idf’: True, ‘vect__ngram_range’: (1, 2)}.


# In[24]:

# Similarly doing grid search for SVM
from sklearn.model_selection import GridSearchCV

parameters_svm = {'vect__ngram_range': [(1, 1), (1, 2)], 'tfidf__use_idf': (True, False),
                  'clf-svm__alpha': (1e-2, 1e-3)}

gs_clf_svm = GridSearchCV(text_clf_svm, parameters_svm, n_jobs=-1)
gs_clf_svm = gs_clf_svm.fit(twenty_train.data, twenty_train.target)

gs_clf_svm.best_score_
gs_clf_svm.best_params_

# In[25]:

# NLTK
# Removing stop words
from sklearn.pipeline import Pipeline

text_clf = Pipeline([('vect', CountVectorizer(stop_words='english')), ('tfidf', TfidfTransformer()),
                     ('clf', MultinomialNB())])

# In[26]:

# Stemming Code

import nltk

nltk.download()

from nltk.stem.snowball import SnowballStemmer

stemmer = SnowballStemmer("english", ignore_stopwords=True)


class StemmedCountVectorizer(CountVectorizer):
    def build_analyzer(self):
        analyzer = super(StemmedCountVectorizer, self).build_analyzer()
        return lambda doc: ([stemmer.stem(w) for w in analyzer(doc)])


stemmed_count_vect = StemmedCountVectorizer(stop_words='english')

text_mnb_stemmed = Pipeline([('vect', stemmed_count_vect), ('tfidf', TfidfTransformer()),
                             ('mnb', MultinomialNB(fit_prior=False))])

text_mnb_stemmed = text_mnb_stemmed.fit(twenty_train.data, twenty_train.target)

predicted_mnb_stemmed = text_mnb_stemmed.predict(twenty_test.data)

np.mean(predicted_mnb_stemmed == twenty_test.target)


# In[ ]:



import pandas as pd

data = pd.read_csv("./10893_sents_shuffled.csv", encoding='latin-1').sample(frac=1).drop_duplicates()
data = data[['LABEL', 'TEXT']].rename(columns={"LABEL": "label", "TEXT": "text"})

data['label'] = '__label__' + data['label'].astype(str)
data.iloc[0:int(len(data) * 0.8)].to_csv('train.csv', sep='\t', index=False, header=False)
data.iloc[int(len(data) * 0.8):int(len(data) * 0.9)].to_csv('test.csv', sep='\t', index=False, header=False)
data.iloc[int(len(data) * 0.9):].to_csv('dev.csv', sep='\t', index=False, header=False);


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	 http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# 형태소분석 기반 BERT를 위한 Tokenization Class
# 수정: joonho.lim
# 일자: 2019-05-23
#
"""Tokenization classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import unicodedata
import os
import logging

# from .file_utils import cached_path
from pytorch_pretrained_bert.file_utils import cached_path
from transformers import BertTokenizer

logger = logging.getLogger(__name__)

PRETRAINED_VOCAB_ARCHIVE_MAP = {
	'bert-base-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-vocab.txt",
	'bert-large-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",
	'bert-base-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-cased-vocab.txt",
	'bert-large-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
	'bert-base-multilingual-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-multilingual-uncased-vocab.txt",
	'bert-base-multilingual-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-multilingual-cased-vocab.txt",
	'bert-base-chinese': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-chinese-vocab.txt",
}
PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP = {
	'bert-base-uncased': 512,
	'bert-large-uncased': 512,
	'bert-base-cased': 512,
	'bert-large-cased': 512,
	'bert-base-multilingual-uncased': 512,
	'bert-base-multilingual-cased': 512,
	'bert-base-chinese': 512,
}
VOCAB_NAME = 'vocab.txt'


def load_vocab(vocab_file):
	"""Loads a vocabulary file into a dictionary."""
	vocab = collections.OrderedDict()
	index = 0
	with open(vocab_file, "r", encoding="utf-8") as reader:
		while True:
			token = reader.readline()
			if not token:
				break
				
			### joonho.lim @ 2019-03-15
			if token.find('n_iters=') == 0 or token.find('max_length=') == 0 :
				continue
			token = token.split('\t')[0]
	  
			token = token.strip()
			vocab[token] = index
			index += 1
	return vocab


def whitespace_tokenize(text):
	"""Runs basic whitespace cleaning and splitting on a peice of text."""
	text = text.strip()
	if not text:
		return []
	tokens = text.split()
	return tokens


class KorBertTokenizer(object):
	"""Runs end-to-end tokenization: punctuation splitting + wordpiece"""

	def __init__(self, vocab_file, do_lower_case=True, max_len=None,
				 never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]")):
		if not os.path.isfile(vocab_file):
			raise ValueError(
				"Can't find a vocabulary file at path '{}'. To load the vocabulary from a Google pretrained "
				"model use `tokenizer = KorBertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`".format(vocab_file))
		self.vocab = load_vocab(vocab_file)
		self.ids_to_tokens = collections.OrderedDict(
			[(ids, tok) for tok, ids in self.vocab.items()])
		self.basic_tokenizer = BasicTokenizer(do_lower_case=do_lower_case,
											  never_split=never_split)
		self.wordpiece_tokenizer = WordpieceTokenizer(vocab=self.vocab)
		self.max_len = max_len if max_len is not None else int(1e12)

	def tokenize(self, text):
		split_tokens = []
		for token in self.basic_tokenizer.tokenize(text):
			### joonho.lim @ 2019-03-15
			token += '_'
			for sub_token in self.wordpiece_tokenizer.tokenize(token):
				split_tokens.append(sub_token)
		return split_tokens

	def convert_tokens_to_ids(self, tokens):
		"""Converts a sequence of tokens into ids using the vocab."""
		ids = []
		for token in tokens:
			ids.append(self.vocab[token])
		if len(ids) > self.max_len:
			raise ValueError(
				"Token indices sequence length is longer than the specified maximum "
				" sequence length for this BERT model ({} > {}). Running this"
				" sequence through BERT will result in indexing errors".format(len(ids), self.max_len)
			)
		return ids

	def convert_ids_to_tokens(self, ids):
		"""Converts a sequence of ids in wordpiece tokens using the vocab."""
		tokens = []
		for i in ids:
			tokens.append(self.ids_to_tokens[i])
		return tokens

	@classmethod
	def from_pretrained(cls, pretrained_model_name, cache_dir=None, *inputs, **kwargs):
		"""
		Instantiate a PreTrainedBertModel from a pre-trained model file.
		Download and cache the pre-trained model file if needed.
		"""
		if pretrained_model_name in PRETRAINED_VOCAB_ARCHIVE_MAP:
			vocab_file = PRETRAINED_VOCAB_ARCHIVE_MAP[pretrained_model_name]
		else:
			vocab_file = pretrained_model_name
		if os.path.isdir(vocab_file):
			vocab_file = os.path.join(vocab_file, VOCAB_NAME)
		# redirect to the cache, if necessary
		try:
			resolved_vocab_file = cached_path(vocab_file, cache_dir=cache_dir)
		except FileNotFoundError:
			logger.error(
				"Model name '{}' was not found in model name list ({}). "
				"We assumed '{}' was a path or url but couldn't find any file "
				"associated to this path or url.".format(
					pretrained_model_name,
					', '.join(PRETRAINED_VOCAB_ARCHIVE_MAP.keys()),
					vocab_file))
			return None
		if resolved_vocab_file == vocab_file:
			logger.info("loading vocabulary file {}".format(vocab_file))
		else:
			logger.info("loading vocabulary file {} from cache at {}".format(
				vocab_file, resolved_vocab_file))
		if pretrained_model_name in PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP:
			# if we're using a pretrained model, ensure the tokenizer wont index sequences longer
			# than the number of positional embeddings
			max_len = PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP[pretrained_model_name]
			kwargs['max_len'] = min(kwargs.get('max_len', int(1e12)), max_len)
		# Instantiate tokenizer.
		tokenizer = cls(resolved_vocab_file, *inputs, **kwargs)
		return tokenizer


class BasicTokenizer(object):
	"""Runs basic tokenization (punctuation splitting, lower casing, etc.)."""

	def __init__(self,
				 do_lower_case=True,
				 never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]")):
		"""Constructs a BasicTokenizer.

		Args:
		  do_lower_case: Whether to lower case the input.
		"""
		self.do_lower_case = do_lower_case
		self.never_split = never_split

	def tokenize(self, text):
		"""Tokenizes a piece of text."""
		text = self._clean_text(text)

		### joonho.lim @ 2019-03-15
		# # # This was added on November 1st, 2018 for the multilingual and Chinese
		# # # models. This is also applied to the English models now, but it doesn't
		# # # matter since the English models were not trained on any Chinese data
		# # # and generally don't have any Chinese data in them (there are Chinese
		# # # characters in the vocabulary because Wikipedia does have some Chinese
		# # # words in the English Wikipedia.).
		# # text = self._tokenize_chinese_chars(text)

		orig_tokens = whitespace_tokenize(text)
		split_tokens = []
		for token in orig_tokens:
			if self.do_lower_case and token not in self.never_split:
				token = token.lower()
				token = self._run_strip_accents(token)
			split_tokens.extend(self._run_split_on_punc(token))

		output_tokens = whitespace_tokenize(" ".join(split_tokens))
		return output_tokens

	def _run_strip_accents(self, text):
		"""Strips accents from a piece of text."""
		text = unicodedata.normalize("NFD", text)
		output = []
		for char in text:
			cat = unicodedata.category(char)
			if cat == "Mn":
				continue
			output.append(char)
		return "".join(output)

	def _run_split_on_punc(self, text):
		"""Splits punctuation on a piece of text."""
		if text in self.never_split:
			return [text]
		chars = list(text)
		i = 0
		start_new_word = True
		output = []
		while i < len(chars):
			char = chars[i]
			if _is_punctuation(char):
				output.append([char])
				start_new_word = True
			else:
				if start_new_word:
					output.append([])
				start_new_word = False
				output[-1].append(char)
			i += 1

		return ["".join(x) for x in output]

	def _tokenize_chinese_chars(self, text):
		"""Adds whitespace around any CJK character."""
		output = []
		for char in text:
			cp = ord(char)
			if self._is_chinese_char(cp):
				output.append(" ")
				output.append(char)
				output.append(" ")
			else:
				output.append(char)
		return "".join(output)

	def _is_chinese_char(self, cp):
		"""Checks whether CP is the codepoint of a CJK character."""
		# This defines a "chinese character" as anything in the CJK Unicode block:
		#   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
		#
		# Note that the CJK Unicode block is NOT all Japanese and Korean characters,
		# despite its name. The modern Korean Hangul alphabet is a different block,
		# as is Japanese Hiragana and Katakana. Those alphabets are used to write
		# space-separated words, so they are not treated specially and handled
		# like the all of the other languages.
		if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
				(cp >= 0x3400 and cp <= 0x4DBF) or  #
				(cp >= 0x20000 and cp <= 0x2A6DF) or  #
				(cp >= 0x2A700 and cp <= 0x2B73F) or  #
				(cp >= 0x2B740 and cp <= 0x2B81F) or  #
				(cp >= 0x2B820 and cp <= 0x2CEAF) or
				(cp >= 0xF900 and cp <= 0xFAFF) or  #
				(cp >= 0x2F800 and cp <= 0x2FA1F)):  #
			return True

		return False

	def _clean_text(self, text):
		"""Performs invalid character removal and whitespace cleanup on text."""
		output = []
		for char in text:
			cp = ord(char)
			if cp == 0 or cp == 0xfffd or _is_control(char):
				continue
			if _is_whitespace(char):
				output.append(" ")
			else:
				output.append(char)
		return "".join(output)


class WordpieceTokenizer(object):
	"""Runs WordPiece tokenization."""

	def __init__(self, vocab, unk_token="[UNK]", max_input_chars_per_word=100):
		self.vocab = vocab
		self.unk_token = unk_token
		self.max_input_chars_per_word = max_input_chars_per_word

	def tokenize(self, text):
		"""Tokenizes a piece of text into its word pieces.

		This uses a greedy longest-match-first algorithm to perform tokenization
		using the given vocabulary.

		For example:
		  input = "unaffable"
		  output = ["un", "##aff", "##able"]

		Args:
		  text: A single token or whitespace separated tokens. This should have
			already been passed through `BasicTokenizer`.

		Returns:
		  A list of wordpiece tokens.
		"""

		output_tokens = []
		for token in whitespace_tokenize(text):
			chars = list(token)
			if len(chars) > self.max_input_chars_per_word:
				output_tokens.append(self.unk_token)
				continue

			is_bad = False
			start = 0
			sub_tokens = []
			while start < len(chars):
				end = len(chars)
				cur_substr = None
				while start < end:
					substr = "".join(chars[start:end])
					### joonho.lim @ 2019-03-15
					# if start > 0:
						# substr = "##" + substr
					if substr in self.vocab:
						cur_substr = substr
						break
					end -= 1
				if cur_substr is None:
					is_bad = True
					break
				sub_tokens.append(cur_substr)
				start = end

			if is_bad:
				output_tokens.append(self.unk_token)
			else:
				output_tokens.extend(sub_tokens)
		return output_tokens


def _is_whitespace(char):
	"""Checks whether `chars` is a whitespace character."""
	# \t, \n, and \r are technically contorl characters but we treat them
	# as whitespace since they are generally considered as such.
	if char == " " or char == "\t" or char == "\n" or char == "\r":
		return True
	cat = unicodedata.category(char)
	if cat == "Zs":
		return True
	return False


def _is_control(char):
	"""Checks whether `chars` is a control character."""
	# These are technically control characters but we count them as whitespace
	# characters.
	if char == "\t" or char == "\n" or char == "\r":
		return False
	cat = unicodedata.category(char)
	if cat.startswith("C"):
		return True
	return False


def _is_punctuation(char):
	### joonho.lim @ 2019-03-15
	return char == ' '
	
	# """Checks whether `chars` is a punctuation character."""
	# cp = ord(char)
	# # We treat all non-letter/number ASCII as punctuation.
	# # Characters such as "^", "$", and "`" are not in the Unicode
	# # Punctuation class but we treat them as punctuation anyways, for
	# # consistency.
	# if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
			# (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
		# return True
	# cat = unicodedata.category(char)
	# if cat.startswith("P"):
		# return True
	# return False



import os

from py_ner.bert_bilstm_crf_ner_train import BertBiLstmCrfTrainer
from py_ner.bert_crf_ner_train import BertCrfTrainer
from shutil import copyfile

data = '../py_ner/data'

def make_dir(directory, parent_dir):
    # Path
    path = os.path.join(parent_dir, directory)

    # Create the directory
    try:
        os.makedirs(path, exist_ok=True)
        print("Directory '%s' created successfully" % directory)
    except OSError as error:
        print("Directory '%s' can not be created")

algorithm = 'bert_gru_crf' #bert_crf, bert_lstm_crf, bert_gru_crf
# we need two mandatory files in this new directory: config.json and ner_to_index.json
if algorithm == 'bert_crf':
    model_dir = 'experiments/base_model_with_crf'
    make_dir(model_dir, "./")
    copyfile('../py_ner/config/config.json', model_dir+"/config.json")
    copyfile('../py_ner/config/ner_to_index.json', model_dir + "/ner_to_index.json")

    trainer = BertCrfTrainer(data_dir=data, model_dir=model_dir)
elif algorithm == 'bert_lstm_crf':
    model_dir = 'experiments/base_model_with_lstm_crf'
    make_dir(model_dir, "./")
    copyfile('../py_ner/config/config.json', model_dir + "/config.json")
    copyfile('../py_ner/config/ner_to_index.json', model_dir + "/ner_to_index.json")

    trainer = BertBiLstmCrfTrainer(data_dir=data, model_dir=model_dir)
elif algorithm == 'bert_gru_crf':
    model_dir = 'experiments/base_model_with_gru_crf'
    make_dir(model_dir, "./")
    copyfile('../py_ner/config/config.json', model_dir + "/config.json")
    copyfile('../py_ner/config/ner_to_index.json', model_dir + "/ner_to_index.json")

    trainer = BertBiLstmCrfTrainer(data_dir=data, model_dir=model_dir)

trainer.data_loading()
trainer.train()


from flair.models import TextClassifier
from flair.data import Sentence

classification_results = open("classification_results.txt", "w")
classifier = TextClassifier.load('./best-model.pt')
filepath = './random500.txt'
with open(filepath) as fp:
    for line in fp:
        #print("Line {}: {}".format(cnt, line.strip()))
        lines = line.split('\t')
        if len(lines) > 2:
            sentence = Sentence(lines[2])
            classifier.predict(sentence)
            print(str(sentence.labels) + "\t" + lines[2])
            classification_results.write(str(sentence.labels) + "\t" + str(lines) + "\n")

fp.close()
classification_results.close()


# coding=utf-8
# Copyright 2018 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Change log
# 12/12/18  Jesse Vig   Adapted to BERT model
# 12/19/18  Jesse Vig   Assorted cleanup. Changed orientation of attention matrices. Updated comments.


"""Module for postprocessing and displaying transformer attentions.

This module is designed to be called from an ipython notebook.
"""

import json
from py_ner.bertviz.attention import get_attention

import os

def show(model, model_type, tokenizer, sentence_a, sentence_b=None):

    if sentence_b:
        vis_html = """
          <span style="user-select:none">
            Attention: <select id="filter">
              <option value="all">All</option>
              <option value="aa">Sentence A -> Sentence A</option>
              <option value="ab">Sentence A -> Sentence B</option>
              <option value="ba">Sentence B -> Sentence A</option>
              <option value="bb">Sentence B -> Sentence B</option>
            </select>
          </span>
          <div id='vis'></div> 
        """
    else:
        vis_html = """
          <div id='vis'></div> 
        """

    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__)))
    vis_js = open(os.path.join(__location__, 'model_view.js')).read()
    attn_data = get_attention(model, model_type, tokenizer, sentence_a, sentence_b)
    params = {
        'attention': attn_data,
        'default_filter': "all"
    }

    with open('bert_visualization.html', 'w') as f:
        f.write(vis_html + '\n')
        f.write("require.config({paths: {d3: '//cdnjs.cloudflare.com/ajax/libs/d3/3.4.8/d3.min',"
                "jquery: '//ajax.googleapis.com/ajax/libs/jquery/2.0.0/jquery.min',}});")

        f.write('window.params = %s' % json.dumps(params) + '\n')
        f.write(vis_js + '\n')




# -*- encoding:utf8 -*-
import pyTextMiner as ptm

dictionary_path='./dict/user_dic.txt'

pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                        ptm.tokenizer.Komoran(),
                        #ptm.tokenizer.WordPos(),
                        ptm.helper.POSFilter('NN*'),
                        ptm.helper.SelectWordOnly(),
                        ptm.helper.StopwordFilter(file='./stopwords/stopwordsKor.txt'),
                        ptm.counter.WordCounter())

corpus = ptm.CorpusFromFile('./data/sampleEng.txt')

#corpus = ptm.CorpusFromFile('Gulliver_Travels.txt')
#pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
#                        ptm.tokenizer.Word(),
#                        ptm.counter.WordCounter())
result = pipeline.processCorpus(corpus)

print(result)
print()

doc_collection = ''
term_counts = {}
for doc in result:
    for sent in doc:
        for _str in sent:
            term_counts[_str[0]] = term_counts.get(_str[0], 0) + int(_str[1])
            freq = range(int(_str[1]))
            co = ''
            for n in freq:
                co +=  ' ' + _str[0]

            doc_collection += ' ' + co
word_freq = []
for key, value in term_counts.items():
    word_freq.append((value,key))

word_freq.sort(reverse=True)
print(word_freq)

f = open("demo_result.txt", "w", encoding='utf8')
for pair in word_freq:
    f.write(pair[1] + '\t' + str(pair[0]) + '\n')
f.close()

from wordcloud import WordCloud

# Read the whole text.

# Generate a word cloud image
wordcloud = WordCloud().generate(doc_collection)

# Display the generated image:
# the matplotlib way:
import matplotlib.pyplot as plt

# Window의 경우 폰트 경로
# font_path = 'C:/Windows/Fonts/malgun.ttf'

#for Mac
#font_path='/Library/Fonts/AppleGothic.ttf'

# lower max_font_size
wordcloud = WordCloud(max_font_size=40,
                      background_color='white',
                      collocations=False)

wordcloud.generate(doc_collection)

plt.figure()
plt.imshow(wordcloud, interpolation="bilinear")
plt.axis("off")
plt.show()



from __future__ import print_function, division
import os
import re
import codecs
import unicodedata
from py_ner.lstm_cnn_crf_utils import create_dico, create_mapping, zero_digits
from py_ner.lstm_cnn_crf_utils import iob2, iob_iobes
import py_ner.lstm_cnn_crf_model as model
import pandas as pd

import string
import random
import numpy as np
import pickle

def unicodeToAscii(s):
    return ''.join(
        c for c in unicodedata.normalize('NFD', s)
        if unicodedata.category(c) != 'Mn'
        and c in string.ascii_letters + " .,;'-"
    )

def load_sentences(path, lower, zeros):
    """
    Load sentences. A line must contain at least a word and its tag.
    Sentences are separated by empty lines.
    """
    sentences = []
    sentence = []
    for line in codecs.open(path, 'r', 'utf-8'):
        line = zero_digits(line.rstrip()) if zeros else line.rstrip()
        if not line:
            if len(sentence) > 0:
                if 'DOCSTART' not in sentence[0][0]:
                    sentences.append(sentence)
                sentence = []
        else:
            word = line.split()
            assert len(word) >= 2
            sentence.append(word)
    if len(sentence) > 0:
        if 'DOCSTART' not in sentence[0][0]:
            sentences.append(sentence)
    return sentences


def update_tag_scheme(sentences, tag_scheme):
    """
    Check and update sentences tagging scheme to IOB2.
    Only IOB1 and IOB2 schemes are accepted.
    """
    for i, s in enumerate(sentences):
        tags = [w[-1] for w in s]
        # Check that tags are given in the IOB format
        if not iob2(tags):
            s_str = '\n'.join(' '.join(w) for w in s)
            raise Exception('Sentences should be given in IOB format! ' +
                            'Please check sentence %i:\n%s' % (i, s_str))
        if tag_scheme == 'iob':
            # If format was IOB1, we convert to IOB2
            for word, new_tag in zip(s, tags):
                word[-1] = new_tag
        elif tag_scheme == 'iobes':
            new_tags = iob_iobes(tags)
            for word, new_tag in zip(s, new_tags):
                word[-1] = new_tag
        else:
            raise Exception('Unknown tagging scheme!')


def word_mapping(sentences, lower):
    """
    Create a dictionary and a mapping of words, sorted by frequency.
    """
    words = [[x[0].lower() if lower else x[0] for x in s] for s in sentences]
    dico = create_dico(words)

    dico['<PAD>'] = 10000001
    dico['<UNK>'] = 10000000
    dico = {k:v for k,v in dico.items() if v>=3}
    word_to_id, id_to_word = create_mapping(dico)

    print("Found %i unique words (%i in total)" % (
        len(dico), sum(len(x) for x in words)
    ))
    return dico, word_to_id, id_to_word


def char_mapping(sentences):
    """
    Create a dictionary and mapping of characters, sorted by frequency.
    """
    chars = ["".join([w[0] for w in s]) for s in sentences]
    dico = create_dico(chars)
    dico['<PAD>'] = 10000000
    # dico[';'] = 0
    char_to_id, id_to_char = create_mapping(dico)
    print("Found %i unique characters" % len(dico))
    return dico, char_to_id, id_to_char


def tag_mapping(sentences):
    """
    Create a dictionary and a mapping of tags, sorted by frequency.
    """
    tags = [[word[-1] for word in s] for s in sentences]
    dico = create_dico(tags)
    dico[model.START_TAG] = -1
    dico[model.STOP_TAG] = -2
    tag_to_id, id_to_tag = create_mapping(dico)
    print("Found %i unique named entity tags" % len(dico))
    return dico, tag_to_id, id_to_tag


def cap_feature(s):
    """
    Capitalization feature:
    0 = low caps
    1 = all caps
    2 = first letter caps
    3 = one capital (not first letter)
    """
    if s.lower() == s:
        return 0
    elif s.upper() == s:
        return 1
    elif s[0].upper() == s[0]:
        return 2
    else:
        return 3


def prepare_sentence(str_words, word_to_id, char_to_id, lower=False):
    """
    Prepare a sentence for evaluation.
    """
    def f(x): return x.lower() if lower else x
    words = [word_to_id[f(w) if f(w) in word_to_id else '<UNK>']
             for w in str_words]
    chars = [[char_to_id[c] for c in w if c in char_to_id]
             for w in str_words]
    caps = [cap_feature(w) for w in str_words]
    return {
        'str_words': str_words,
        'words': words,
        'chars': chars,
        'caps': caps
    }


def prepare_dataset(sentences, word_to_id, char_to_id, tag_to_id, lower=True):
    """
    Prepare the dataset. Return a list of lists of dictionaries containing:
        - word indexes
        - word char indexes
        - tag indexes
    """
    def f(x): return x.lower() if lower else x
    data = []
    for s in sentences:
        str_words = [w[0] for w in s]
        words = [word_to_id[f(w) if f(w) in word_to_id else '<UNK>']
                 for w in str_words]
        # Skip characters that are not in the training set
        chars = [[char_to_id[c] for c in w if c in char_to_id]
                 for w in str_words]
        caps = [cap_feature(w) for w in str_words]
        tags = [tag_to_id[w[-1]] for w in s]
        data.append({
            'str_words': str_words,
            'words': words,
            'chars': chars,
            'caps': caps,
            'tags': tags,
        })
    return data


def augment_with_pretrained(dictionary, ext_emb_path, words):
    """
    Augment the dictionary with words that have a pretrained embedding.
    If `words` is None, we add every word that has a pretrained embedding
    to the dictionary, otherwise, we only add the words that are given by
    `words` (typically the words in the development and test sets.)
    """
    print('Loading pretrained embeddings from %s...' % ext_emb_path)
    assert os.path.isfile(ext_emb_path)

    # Load pretrained embeddings from file
    pretrained = set([
        line.rstrip().split()[0].strip()
        for line in codecs.open(ext_emb_path, 'r', 'utf-8')
        if len(ext_emb_path) > 0
    ])

    # We either add every word in the pretrained file,
    # or only words given in the `words` list to which
    # we can assign a pretrained embedding
    if words is None:
        for word in pretrained:
            if word not in dictionary:
                dictionary[word] = 0
    else:
        for word in words:
            if any(x in pretrained for x in [
                word,
                word.lower(),
                re.sub('\d', '0', word.lower())
            ]) and word not in dictionary:
                dictionary[word] = 0

    word_to_id, id_to_word = create_mapping(dictionary)
    return dictionary, word_to_id, id_to_word


def pad_seq(seq, max_length, PAD_token=0):
    # add pads
    seq += [PAD_token for i in range(max_length - len(seq))]
    return seq

def get_batch(start, batch_size, datas, singletons=[]):
    input_seqs = []
    target_seqs = []
    chars2_seqs = []

    for data in datas[start:start+batch_size]:
        # pair is chosen from pairs randomly
        words = []
        for word in data['words']:
            if word in singletons and np.random.uniform() < 0.5:
                words.append(1)
            else:
                words.append(word)
        input_seqs.append(data['words'])
        target_seqs.append(data['tags'])
        chars2_seqs.append(data['chars'])

    if input_seqs == []:
        return [], [], [], [], [], []
    seq_pairs = sorted(zip(input_seqs, target_seqs, chars2_seqs), key=lambda p: len(p[0]), reverse=True)
    input_seqs, target_seqs, chars2_seqs = zip(*seq_pairs)

    chars2_seqs_lengths = []
    chars2_seqs_padded = []
    for chars2 in chars2_seqs:
        chars2_lengths = [len(c) for c in chars2]
        chars2_padded = [pad_seq(c, max(chars2_lengths)) for c in chars2]
        chars2_seqs_padded.append(chars2_padded)
        chars2_seqs_lengths.append(chars2_lengths)

    input_lengths = [len(s) for s in input_seqs]
    # input_padded is batch * max_length
    input_padded = [pad_seq(s, max(input_lengths)) for s in input_seqs]
    target_lengths = [len(s) for s in target_seqs]
    assert target_lengths == input_lengths
    # target_padded is batch * max_length
    target_padded = [pad_seq(s, max(target_lengths)) for s in target_seqs]

    # var is max_length * batch_size
    # input_var = Variable(torch.LongTensor(input_padded)).transpose(0, 1)
    # target_var = Variable(torch.LongTensor(target_padded)).transpose(0, 1)
    #
    # if use_gpu:
    #     input_var = input_var.cuda()
    #     target_var = target_var.cuda()

    return input_padded, input_lengths, target_padded, target_lengths, chars2_seqs_padded, chars2_seqs_lengths


def random_batch(batch_size, train_data, singletons=[]):
    input_seqs = []
    target_seqs = []
    chars2_seqs = []

    for i in range(batch_size):
        # pair is chosen from pairs randomly
        data = random.choice(train_data)
        words = []
        for word in data['words']:
            if word in singletons and np.random.uniform() < 0.5:
                words.append(1)
            else:
                words.append(word)
        input_seqs.append(data['words'])
        target_seqs.append(data['tags'])
        chars2_seqs.append(data['chars'])

    seq_pairs = sorted(zip(input_seqs, target_seqs, chars2_seqs), key=lambda p: len(p[0]), reverse=True)
    input_seqs, target_seqs, chars2_seqs = zip(*seq_pairs)

    chars2_seqs_lengths = []
    chars2_seqs_padded = []
    for chars2 in chars2_seqs:
        chars2_lengths = [len(c) for c in chars2]
        chars2_padded = [pad_seq(c, max(chars2_lengths)) for c in chars2]
        chars2_seqs_padded.append(chars2_padded)
        chars2_seqs_lengths.append(chars2_lengths)

    input_lengths = [len(s) for s in input_seqs]
    # input_padded is batch * max_length
    input_padded = [pad_seq(s, max(input_lengths)) for s in input_seqs]
    target_lengths = [len(s) for s in target_seqs]
    assert target_lengths == input_lengths
    # target_padded is batch * max_length
    target_padded = [pad_seq(s, max(target_lengths)) for s in target_seqs]

    # var is max_length * batch_size
    # input_var = Variable(torch.LongTensor(input_padded)).transpose(0, 1)
    # target_var = Variable(torch.LongTensor(target_padded)).transpose(0, 1)
    #
    # if use_gpu:
    #     input_var = input_var.cuda()
    #     target_var = target_var.cuda()

    return input_padded, input_lengths, target_padded, target_lengths, chars2_seqs_padded, chars2_seqs_lengths

def convert_to_df(text_file):
    dframe = pd.read_csv(text_file, encoding="ISO-8859-1", error_bad_lines=False)
    dataset = dframe.drop(['Unnamed: 0', 'lemma', 'next-lemma', 'next-next-lemma', 'next-next-pos',
                           'next-next-shape', 'next-next-word', 'next-pos', 'next-shape',
                           'next-word', 'prev-iob', 'prev-lemma', 'prev-pos',
                           'prev-prev-iob', 'prev-prev-lemma', 'prev-prev-pos', 'prev-prev-shape',
                           'prev-prev-word', 'prev-shape', 'prev-word', 'shape'], axis=1)

    return dataset

def convert_to_df_for_ko_ner(sentences):
    pd.set_option('display.max_columns', None)
    document_df = pd.DataFrame()

    max = 500;
    for i, sent in enumerate(sentences):
        for ej_idx, token, postag, label in sent:
            document_df = document_df.append(pd.Series([int(i), token, postag, label]), ignore_index=True)
        i += 1

        if i >= max:
            break

    document_df.columns = ['sentence_idx', 'word', 'pos', 'tag']
    document_df = document_df.set_index('sentence_idx')
    tag_names = document_df['tag'].unique()

    return document_df, tag_names

def read_file(file_name):
    sents = []
    with open(file_name,'r',encoding='utf-8') as f:
        lines = f.readlines()
        for idx,l in enumerate(lines) :
            if l[0]==';' and lines[idx+1][0]=='$':
                this_sent = []
            elif l[0]=='$' and lines[idx-1][0]==';':
                continue
            elif l[0]=='\n':
                sents.append(this_sent)
            else :
                this_sent.append(tuple(l.split()))
    return sents

def word2features(sent, i):
    last_p = sent[i][0]
    mor_idx = sent[i][1]
    word = sent[i][3]
    postag = sent[i][4]
    features = [
        'bias',
        'word=' + word,
        'word[:1]=' + word[:1],
        'word[:2]=' + word[:2],
        'word[-3:]=' + word[-3:],
        'word[-2:]=' + word[-2:],
        'word.isdigit=%s' % word.isdigit(),
        'postag=' + postag,
        'postag[:2]=' + postag[:2],
        'last_p=' + last_p,
        'mor_i=' + mor_idx,
        'len='+str(len(word))
    ]
    if i > 0:
        word1 = sent[i-1][3]
        postag1 = sent[i-1][4]
        features.extend([
            '-1:word=' + word1,
            '-1:postag=' + postag1,
            '-1:postag[:2]=' + postag1[:2],
        ])
    else:
        features.append('BOS')

    if i < len(sent) - 1:
        word1 = sent[i+1][3]
        postag1 = sent[i+1][4]
        features.extend([
            '+1:word=' + word1,
            '+1:postag=' + postag1,
            '+1:postag[:2]=' + postag1[:2],
        ])
    else:
        features.append('EOS')

    return features

def add_features_to_sent(sent):
    new_sent = []
    this_ej = []
    ej_idx = 1
    morph_idx = 1
    for w in sent:
        if int(w[0]) > ej_idx:
            assert(len(this_ej)>0)
            last_particle = '-'
            if this_ej[-1][3][0]=='J':
                last_particle=this_ej[-1][2]+'/'+this_ej[-1][3]
            this_ej = [(last_particle,)+m for m in this_ej]
            new_sent += this_ej
            this_ej = []
            ej_idx += 1
            morph_idx = 1
        this_ej.append((str(morph_idx),)+w)
        morph_idx += 1
    if len(this_ej) > 0:
        last_particle = '-'
        if this_ej[-1][3][0] == 'J':
            last_particle = this_ej[-1][2] + '/' + this_ej[-1][3]
        this_ej = [(last_particle,) + m for m in this_ej]
        new_sent += this_ej
    return new_sent

def sent2features(sent):
    sent = add_features_to_sent(sent)
    return [word2features(sent, i) for i in range(len(sent))]

def sent2labels(sent):
    return [label for ej_idx, token, postag, label in sent]

def sent2tokens(sent):
    return [token for ej_idx, token, postag, label in sent]


class SentenceGetter(object):
    def __init__(self, dataset):
        self.n_sent = 1
        self.dataset = dataset
        self.empty = False
        agg_func = lambda s: [(w, p, t) for w, p, t in zip(s["word"].values.tolist(),
                                                           s['pos'].values.tolist(),
                                                           s["tag"].values.tolist())]
        self.grouped = self.dataset.groupby("sentence_idx").apply(agg_func)
        self.sentences = [s for s in self.grouped]

    def get_next(self):
        try:
            s = self.grouped["Sentence: {}".format(self.n_sent)]
            self.n_sent += 1
            return s
        except:
            return None



import argparse
import flair
from docx import Document
from flair.data import Sentence
from flair.models import SequenceTagger
from flair.embeddings import WordEmbeddings, FlairEmbeddings, StackedEmbeddings, DocumentPoolEmbeddings, BertEmbeddings, ELMoEmbeddings, OpenAIGPTEmbeddings, TransformerXLEmbeddings, XLNetEmbeddings
import torch
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.metrics import jaccard_similarity_score
#import numpy as np

import sys
import numpy as np
from itertools import islice
from collections import deque
import matplotlib
#import umap ##only necessary for weird experiments
import matplotlib.pyplot as plt
#import seaborn as sns ##only necessary for weird experiments
from mpl_toolkits.mplot3d import proj3d
import matplotlib.cm as cm
from torch.nn import CosineSimilarity
from sty import fg, bg, ef, rs, RgbFg
from sklearn.preprocessing import MinMaxScaler
import syntok.segmenter as segmenter
from ansi2html import Ansi2HTMLConverter
import py_summarization.text_graph

conv = Ansi2HTMLConverter()

document = Document() ## Create a python-docx document
cos = CosineSimilarity(dim=1, eps=1e-6)



class bcolors: #For converting the ANSI string to HTML - Sty is not supported well :(
    HIGHLIGHT = '\33[43m'
    END = '\033[0m'


granularity_level = "Sent" #"Word" "Sent" "Paragraph"
dynamic = False  ##Controls if we highlight the more important words more or not
graph = False ###ONLY WORKS WITH "Word" granularity_level
word_doc = True
html = True
word_window_size = 10 ##This is really doubled since it's Bi-directional. Only matters for word level granularity
highlight_color_intensity = 175 # try values between 25 and 200
doc_embeddings = []
scores = []

stacked_embeddings = DocumentPoolEmbeddings([
                                        #WordEmbeddings('en'),
                                        XLNetEmbeddings('xlnet-large-cased', use_scalar_mix = 'True')
                                        #WordEmbeddings('glove'),
                                        #WordEmbeddings('extvec'),#ELMoEmbeddings('original'),
                                        #BertEmbeddings('bert-base-cased'),
                                        #FlairEmbeddings('news-forward-fast'),
                                        #FlairEmbeddings('news-backward-fast'),
                                        #OpenAIGPTEmbeddings()
                                        #TransformerXLEmbeddings()
                                        ]) #, mode='max')

def arg_parser():
  parser = argparse.ArgumentParser()
  parser.add_argument("--card_file_path", "-c", type=str,
                      help="Path to file for file text")
  args = parser.parse_args()
  return args

def set_card_text(card_path=None):
  card = ""
  if card_path:
    with open(card_path, "r") as in_file:
      card = in_file.read()
  else:
    print("Input the Card Text, press Ctrl-D to end text entry")
    card = sys.stdin.read()#input("Input the Card Text: ")
  return card


def set_card(card_path=None):
    card = set_card_text(card_path)
    card_tag = input("Input the card_tag, or a -1 to summarize in-terms of the card itself: ")
    card = str(card)
    if str(card_tag) == "-1": #This will not work with large documents when bert is enabled
        card_tag = Sentence(str(card))
        tag_str = ""
    else:
        tag_str = str(card_tag)
        card_tag = Sentence(str(card_tag))
    return card, card_tag, tag_str


def create_ngram(num_context, card, card_tag):
    #card = set_card()
    card_words_org = card.split()
    ngram_list = []
    #print(len(card_words_org))

    for i in range(0, len(card_words_org)):
        # Get the sliding window.
        lower_bound = i - num_context if i - num_context > 0 else 0
        upper_bound = i + num_context if i + num_context < len(card_words_org) else len(card_words_org) - 1
        new_word = card_words_org[lower_bound : upper_bound]
        print(new_word)

        # Join the window.
        new_string = " ".join(new_word)
        if new_string == "":
          new_string = " "
        ngram_list.append(Sentence(new_string))

    return ngram_list, card_words_org


def embed(card_tag, card_as_sentence, card_words, card_words_org):
    stacked_embeddings.embed(card_tag)
    #stacked_embeddings.embed(card_as_sentence)
    #print(card_as_sentence.get_embedding().reshape(1,-1))
    word_list = []
    token_removed_ct = 0
    card_tag_emb = card_tag.get_embedding()
    if granularity_level == "Word":
        for word, count in zip(card_words_org, range(0, len(card_words_org))):
            n_gram_word = card_words[count]
            stacked_embeddings.embed(n_gram_word)
            n_gram_emb = n_gram_word.get_embedding()
            if graph:
                doc_embeddings.append(n_gram_emb.cpu().detach().numpy())
            word_sim = cos(card_tag_emb.reshape(1,-1), n_gram_emb.reshape(1, -1))
            word_tup = (card_words_org[count], word_sim) #card_words_org[count]
            word_list.append(word_tup)
        if graph:
            doc_embeddings.append(card_tag_emb.cpu().detach().numpy())
        print(len(word_list))
        print(len(card_words))
        print(len(card_words_org))
    else:
        for sentence in card_as_sentence:
            set_obj = Sentence(sentence)
            stacked_embeddings.embed(set_obj)
            sentence_emb = set_obj.get_embedding()
            word_sim = cos(card_tag_emb.reshape(1,-1), sentence_emb.reshape(1, -1))
            sentence_tup = (sentence, word_sim)
            word_list.append(sentence_tup)
    return word_list


def summarize(word_list):
    return [sum_word[1] for sum_word in word_list]

def get_sim_scores(word_list):
    return [float(sum_word[1]) for sum_word in word_list]

def run_loop(context, card, card_tag):
    list_of_sentences = []
    list_of_paragraphs = []
    if granularity_level == "Sent":
        for paragraph in segmenter.analyze(card):
            for sentence in paragraph: ## sentence level summarization
                set_str = ""
                for token in sentence:
                    set_str += token.spacing
                    set_str += token.value
                list_of_sentences.append(set_str)
        word_list = embed(card_tag, list_of_sentences, 0, 0)
    elif granularity_level == "Paragraph":
        for paragraph in segmenter.analyze(card):
            set_str = ""
            for sentence in paragraph: ## sentence level summarization
                #set_str = ""
                for token in sentence:
                    set_str += token.spacing
                    set_str += token.value
            list_of_paragraphs.append(set_str)
        word_list = embed(card_tag, list_of_paragraphs, 0, 0)
    elif granularity_level == "Word":
        card_as_sentence = Sentence(card)
        card_words, card_words_org = create_ngram(context, card, card_tag)
        word_list = embed(card_tag, card_as_sentence, card_words, card_words_org)
    #print(word_list)
    return word_list

def parse_word_val_list(word_list, h, n, val_list):
    sum_str = ""
    html_str = ""
    removed_str = ""
    token_removed_ct = 0
    to_highlight = np.asarray([i for i in val_list if i > h])
    scaler = MinMaxScaler(feature_range = (20, 255))
    scaler.fit(to_highlight.reshape(-1, 1))
    for sum_word in word_list:
        prev_word = "n"
        if float(sum_word[1]) > h:
            if dynamic:
                bgnum = int(scaler.transform(sum_word[1].cpu().detach().numpy().reshape(-1, 1)))
            else:
                bgnum = highlight_color_intensity
            html_str += ef.u + bcolors.HIGHLIGHT + sum_word[0] + " "
            sum_str += ef.u + bg(bgnum, bgnum, 0)
            runner = par.add_run(sum_word[0] + " ")
            runner.underline = True
            runner.bold = True
            sum_str += " ".join([str(sum_word[0]), rs.all])
        elif float(sum_word[1]) > n:
            html_str += ef.u + sum_word[0] + " "
            sum_str += ef.u
            runner = par.add_run(sum_word[0] + " ")
            runner.underline = True
            sum_str += " ".join([str(sum_word[0]), rs.all])
        else:
            html_str += sum_word[0] + " " + bcolors.END
            token_removed_ct += 1
            sum_str += str(sum_word[0])
            runner = par.add_run(sum_word[0] + " ")
            sum_str += " "
            removed_str += str(sum_word[0])
            removed_str += " "
    return sum_str, removed_str, token_removed_ct, html_str

def new_sum(word_list, underline_p, highlight_p):
    val_list = summarize(word_list)
    th_underline = np.percentile(val_list, float(underline_p))
    th_highlight = np.percentile(val_list, float(highlight_p))
    sum_str, removed_str, token_removed_ct, t_html_str = parse_word_val_list(word_list, th_highlight, th_underline, val_list)


    print("CARD: ")
    print(card)
    print("CARD_TAG :")
    print(card_tag)
    print("GENERATED SUMMARY: ")
    print(sum_str)
    print("tokens removed:" + " " + str(token_removed_ct))
    #print(conv.convert(sum_str))
    if html:
        #print(conv.convert(t_html_str))
        with open("cx_html.html", "a") as hf:
            hf.write(conv.convert(t_html_str))

for i in range(0, 1000):
    args = arg_parser()
    cont = input("Do you want to continue? type y for yes else for stop!")
    if cont == "y":
        card, card_tag, tag_str = set_card(args.card_file_path)
        head = document.add_heading(tag_str, level=1)
        underline_p = input("what percentile should be underlined? (numbers from 1 to 99 acceptable, a 90 means only the top 10% is underlined")
        highlight_p = input("what percentile to highlight? (should be higher than the previous value)")
        card_auth = input("what's the card author and date?")
        card_cite = input("what's the citation?")
        a_par = document.add_paragraph()
        a_par.add_run(card_auth).bold = True
        c_par = document.add_paragraph(card_cite)
        par = document.add_paragraph()
        word_list = run_loop(word_window_size, card, card_tag)
        if graph:
            #plt.figure(figsize=(15,1))
            #sns.heatmap(doc_embeddings[0:20], robust = True, linewidths=0.7, cmap="PiYG")
            #plt.show()
            gph = np.array([doc_embeddings[k][0:3].tolist() for k in range(0, len(doc_embeddings))])
            word_list.append(('CARD_TAG', 1.0))
            scores = get_sim_scores(word_list)
            #scores.append(1.0) #to account for the query string)
            text_graph.visualize3DData(gph, scores, word_list)

            '''
            #umap_emb = reducer.fit_transform(doc_embeddings) ##if we want to fit with the card tag, maybe not as interesting
            gph = np.array([doc_embeddings[k][0:3].tolist() for k in range(0, len(doc_embeddings))])
            #umap_emb = reducer.fit_transform((doc_embeddings[:-1]))
            #t_emb = reducer.transform(doc_embeddings[-1].reshape(1, -1))
            #n_umap_emb = np.append(umap_emb, t_emb, axis=0)
            scores = get_sim_scores(word_list)
            scores.append(1.0) #to account for the query string)
            '''
        new_sum(word_list, underline_p, highlight_p)
    else:
        if word_doc:
            document.save('test_sum.docx')
        break



import json
import pickle
import argparse
import torch
from py_ner.kobert.pytorch_kobert import get_pytorch_kobert_model
from py_ner.kobert.utils import get_tokenizer
from py_ner.model.net import KobertSequenceFeatureExtractor, KobertCRF, KobertBiLSTMCRF, KobertBiGRUCRF
from gluonnlp.data import SentencepieceTokenizer
from py_ner.data_utils.utils import Config
from py_ner.data_utils.vocab_tokenizer import Tokenizer
from py_ner.data_utils.pad_sequence import keras_pad_fn
from pathlib import Path

class BertCRFNERPredictor:
    def __init__(self, model_dir=''):
        self.model_dir = model_dir
        self.model_config = Config(json_path=self.model_dir + '/config.json')
        self.tokenizer = None
        self.model = None
        self.decoder_from_res = None
        self.model_name = ''

    def load_model(self, model_name='bert_crf',
                   tokenizer_path='./ptr_lm_model/tokenizer_78b3253a26.model',
                   checkpoint_file='./experiments/base_model_with_crf_val/best-epoch-9-step-750-acc-0.980.bin'):
        # Vocab & Tokenizer
        # tok_path = get_tokenizer() # ./tokenizer_78b3253a26.model
        #tok_path = "./ptr_lm_model/tokenizer_78b3253a26.model"
        tok_path = tokenizer_path
        ptr_tokenizer = SentencepieceTokenizer(tok_path)

        # load vocab & tokenizer
        with open(self.model_dir + "/vocab.pkl", 'rb') as f:
            vocab = pickle.load(f)

        self.tokenizer = Tokenizer(vocab=vocab, split_fn=ptr_tokenizer, pad_fn=keras_pad_fn, maxlen=self.model_config.maxlen)

        # load ner_to_index.json
        with open(self.model_dir + "/ner_to_index.json", 'rb') as f:
            ner_to_index = json.load(f)
            index_to_ner = {v: k for k, v in ner_to_index.items()}

        # Model
        # model = KobertSequenceFeatureExtractor(config=self.model_config, num_classes=len(ner_to_index))
        if model_name == 'bert_lstm_crf':
            self.model = KobertBiLSTMCRF(config=self.model_config, num_classes=len(ner_to_index), vocab=vocab)
            checkpoint = torch.load(checkpoint_file,
                                    map_location=torch.device('cpu'))
        elif model_name == 'bert_gru_crf':
            self.model = KobertBiGRUCRF(config=self.model_config, num_classes=len(ner_to_index), vocab=vocab)
            checkpoint = torch.load(checkpoint_file,
                                    map_location=torch.device('cpu'))
        else:
            self.model = KobertCRF(config=self.model_config, num_classes=len(ner_to_index), vocab=vocab)
            checkpoint = torch.load(checkpoint_file,
                                    map_location=torch.device('cpu'))
        self.model_name = model_name

        # load
        model_dict = self.model.state_dict()
        # checkpoint = torch.load("./experiments/base_model/best-epoch-9-step-600-acc-0.845.bin", map_location=torch.device('cpu'))
        # checkpoint = torch.load("./experiments/base_model_with_crf/best-epoch-16-step-1500-acc-0.993.bin", map_location=torch.device('cpu'))
        #checkpoint = torch.load("./experiments/base_model_with_crf_val/best-epoch-12-step-1000-acc-0.960.bin", map_location=torch.device('cpu'))
        # checkpoint = torch.load("./experiments/base_model_with_bilstm_crf/best-epoch-15-step-2750-acc-0.992.bin", map_location=torch.device('cpu'))
        # checkpoint = torch.load("./experiments/base_model_with_bigru_crf/model-epoch-18-step-3250-acc-0.997.bin", map_location=torch.device('cpu'))

        convert_keys = {}
        for k, v in checkpoint['model_state_dict'].items():
            new_key_name = k.replace("module.", '')
            if new_key_name not in model_dict:
                print("{} is not int model_dict".format(new_key_name))
                continue
            convert_keys[new_key_name] = v

        self.model.load_state_dict(convert_keys)
        self.model.eval()
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        # n_gpu = torch.cuda.device_count()
        # if n_gpu > 1:
        #     model = torch.nn.DataParallel(model)
        self.model.to(device)

        self.decoder_from_res = DecoderFromNamedEntitySequence(tokenizer=self.tokenizer, index_to_ner=index_to_ner)

    def predict(self, input_text):
        print("문장을 입력하세요: ")
        list_of_input_ids = self.tokenizer.list_of_string_to_list_of_cls_sep_token_ids([input_text])
        x_input = torch.tensor(list_of_input_ids).long()

        ## for bert alone
        # y_pred = model(x_input)
        # list_of_pred_ids = y_pred.max(dim=-1)[1].tolist()

        if self.model_name == 'bert_lstm_crf' or self.model_name == 'bert_gru_crf':
            ## for bert bilstm crf & bert bigru crf
            list_of_pred_ids = self.model(x_input, using_pack_sequence=False)
        else:
            ## for bert crf
            list_of_pred_ids = self.model(x_input)

        list_of_ner_word, decoding_ner_sentence = self.decoder_from_res(list_of_input_ids=list_of_input_ids, list_of_pred_ids=list_of_pred_ids)
        print("list_of_ner_word:", list_of_ner_word)
        #print("decoding_ner_sentence:", decoding_ner_sentence)
        return decoding_ner_sentence

class DecoderFromNamedEntitySequence():
    def __init__(self, tokenizer, index_to_ner):
        self.tokenizer = tokenizer
        self.index_to_ner = index_to_ner

    def __call__(self, list_of_input_ids, list_of_pred_ids):
        input_token = self.tokenizer.decode_token_ids(list_of_input_ids)[0]
        print(input_token)
        print(self.index_to_ner[6])

        _list = list_of_pred_ids[0].tolist()
        pred_ner_tag = [self.index_to_ner[pred_id] for pred_id in _list[0]]

        print("len: {}, input_token:{}".format(len(input_token), input_token))
        print("len: {}, pred_ner_tag:{}".format(len(pred_ner_tag), pred_ner_tag))

        # ----------------------------- parsing list_of_ner_word ----------------------------- #
        list_of_ner_word = []
        entity_word, entity_tag, prev_entity_tag = "", "", ""
        for i, pred_ner_tag_str in enumerate(pred_ner_tag):
            if "B-" in pred_ner_tag_str:
                entity_tag = pred_ner_tag_str[-3:]

                if prev_entity_tag != entity_tag and prev_entity_tag != "":
                    list_of_ner_word.append({"word": entity_word.replace("▁", " "), "tag": prev_entity_tag, "prob": None})

                entity_word = input_token[i]
                prev_entity_tag = entity_tag
            elif "I-"+entity_tag in pred_ner_tag_str:
                entity_word += input_token[i]
            else:
                if entity_word != "" and entity_tag != "":
                    list_of_ner_word.append({"word":entity_word.replace("▁", " "), "tag":entity_tag, "prob":None})
                entity_word, entity_tag, prev_entity_tag = "", "", ""

        # ----------------------------- parsing decoding_ner_sentence ----------------------------- #
        decoding_ner_sentence = ""
        is_prev_entity = False
        prev_entity_tag = ""
        is_there_B_before_I = False

        for token_str, pred_ner_tag_str in zip(input_token, pred_ner_tag):
            token_str = token_str.replace('▁', ' ')  # '▁' 토큰을 띄어쓰기로 교체

            if 'B-' in pred_ner_tag_str:
                if is_prev_entity is True:
                    decoding_ner_sentence += ':' + prev_entity_tag+ '>'

                if token_str[0] == ' ':
                    token_str = list(token_str)
                    token_str[0] = ' <'
                    token_str = ''.join(token_str)
                    decoding_ner_sentence += token_str
                else:
                    decoding_ner_sentence += '<' + token_str
                is_prev_entity = True
                prev_entity_tag = pred_ner_tag_str[-3:] # 첫번째 예측을 기준으로 하겠음
                is_there_B_before_I = True

            elif 'I-' in pred_ner_tag_str:
                decoding_ner_sentence += token_str

                if is_there_B_before_I is True: # I가 나오기전에 B가 있어야하도록 체크
                    is_prev_entity = True
            else:
                if is_prev_entity is True:
                    decoding_ner_sentence += ':' + prev_entity_tag+ '>' + token_str
                    is_prev_entity = False
                    is_there_B_before_I = False
                else:
                    decoding_ner_sentence += token_str

        return list_of_ner_word, decoding_ner_sentence

if __name__ == '__main__':
    model_dir = './experiments/base_model_with_crf_val'
    predictor = BertCRFNERPredictor(model_dir)

    model_name = 'bert_lstm_crf'
    tokenizer_path = "./ptr_lm_model/tokenizer_78b3253a26.model"
    predictor.load_model(model_name, tokenizer_path)

    text = '김대중 대통령은 노벨평화상을 받으러 스웨덴으로 출국해서 5박6일 동안 스웨덴에 머물며 대한민국의 위상을 높였다.'
    predictor.predict(text)



import pycrfsuite

def read_file(file_name):
    sents = []
    with open(file_name,'r',encoding='utf-8') as f:
        lines = f.readlines()
        for idx,l in enumerate(lines) :
            if l[0]==';' and lines[idx+1][0]=='$':
                this_sent = []
            elif l[0]=='$' and lines[idx-1][0]==';':
                continue
            elif l[0]=='\n':
                sents.append(this_sent)
            else :
                this_sent.append(tuple(l.split()))
    return sents

train_sents = read_file('data/train.txt')
test_sents = read_file('data/test.txt')

def word2features(sent, i):
    last_p = sent[i][0]
    mor_idx = sent[i][1]
    word = sent[i][3]
    postag = sent[i][4]
    features = [
        'bias',
        'word=' + word,
        'word[:1]=' + word[:1],
        'word[:2]=' + word[:2],
        'word[-3:]=' + word[-3:],
        'word[-2:]=' + word[-2:],
        'word.isdigit=%s' % word.isdigit(),
        'postag=' + postag,
        'postag[:2]=' + postag[:2],
        'last_p=' + last_p,
        'mor_i=' + mor_idx,
        'len='+str(len(word))
    ]
    if i > 0:
        word1 = sent[i-1][3]
        postag1 = sent[i-1][4]
        features.extend([
            '-1:word=' + word1,
            '-1:postag=' + postag1,
            '-1:postag[:2]=' + postag1[:2],
        ])
    else:
        features.append('BOS')

    if i < len(sent) - 1:
        word1 = sent[i+1][3]
        postag1 = sent[i+1][4]
        features.extend([
            '+1:word=' + word1,
            '+1:postag=' + postag1,
            '+1:postag[:2]=' + postag1[:2],
        ])
    else:
        features.append('EOS')

    return features

def add_features_to_sent(sent):
    new_sent = []
    this_ej = []
    ej_idx = 1
    morph_idx = 1
    for w in sent:
        if int(w[0]) > ej_idx:
            assert(len(this_ej)>0)
            last_particle = '-'
            if this_ej[-1][3][0]=='J':
                last_particle=this_ej[-1][2]+'/'+this_ej[-1][3]
            this_ej = [(last_particle,)+m for m in this_ej]
            new_sent += this_ej
            this_ej = []
            ej_idx += 1
            morph_idx = 1
        this_ej.append((str(morph_idx),)+w)
        morph_idx += 1
    if len(this_ej) > 0:
        last_particle = '-'
        if this_ej[-1][3][0] == 'J':
            last_particle = this_ej[-1][2] + '/' + this_ej[-1][3]
        this_ej = [(last_particle,) + m for m in this_ej]
        new_sent += this_ej
    return new_sent

def sent2features(sent):
    sent = add_features_to_sent(sent)
    return [word2features(sent, i) for i in range(len(sent))]

def sent2labels(sent):
    return [label for ej_idx, token, postag, label in sent]


def sent2tokens(sent):
    return [token for ej_idx, token, postag, label in sent]

#data에서 자질 추출
X_train = [sent2features(s) for s in train_sents]
y_train = [sent2labels(s) for s in train_sents]

X_test = [sent2features(s) for s in test_sents]
y_test = [sent2labels(s) for s in test_sents]

#중간 출력
print(X_train[0][2])

#모델 학습
trainer = pycrfsuite.Trainer(verbose=False)

for xseq, yseq in zip(X_train, y_train):
    trainer.append(xseq, yseq)

trainer.set_params({
    'c1': 1.0,   # coefficient for L1 penalty
    'c2': 1e-3,  # coefficient for L2 penalty
    'max_iterations': 50,  # stop earlier

    # include transitions that are possible, but not observed
    'feature.possible_transitions': True
})

trainer.train('model/kor.crfsuite')
#모델 로드
tagger = pycrfsuite.Tagger()
tagger.open('model/kor.crfsuite')
def make_tag_idx(ans_seq):
    B_num = 0
    all_answer_start = []
    all_answer_end = []
    all_answer_tag = []

    for sent in ans_seq:
        start_idx = []
        end_idx = []
        tag_set = []
        tag_num = 0
        flag = 0

        for tag in sent:
            try:
                if flag == 1 and tag[0] != 'I':
                    end_idx.append(tag_num - 1)
                    flag = 0
            except:
                print()

            if tag[0] == 'B':
                B_num = B_num + 1
                start_idx.append(tag_num)
                tag_set.append(tag[2:4])
                flag = 1
            tag_num = tag_num + 1

        if flag == 1:
            end_idx.append(tag_num - 1)

        all_answer_start.append(start_idx)
        all_answer_end.append(end_idx)
        all_answer_tag.append(tag_set)

    return (all_answer_start, all_answer_end, all_answer_tag, B_num)

def eval(pred_seq, ans_seq):  # changed

    TP = 0

    (all_answer_start, all_answer_end, all_answer_tag, answer_num) = make_tag_idx(ans_seq)
    (all_pred_start, all_pred_end, all_pred_tag, pred_num) = make_tag_idx(pred_seq)

    for i in range(0, len(ans_seq)):
        for j in range(0, len(all_pred_start[i])):
            for k in range(0, len(all_answer_start[i])):
                if all_pred_start[i][j] == all_answer_start[i][k] and all_pred_end[i][j] == all_answer_end[i][k] and \
                                all_pred_tag[i][j] == all_answer_tag[i][k]: TP = TP + 1

    return (TP, pred_num, answer_num)

def write_prediction(file_name, sents, preds):
    with open(file_name,'w', encoding='utf-8') as f:
        for s, s_p in zip(sents,preds):
            f.write(";\n$\n")
            for t, p in zip(s,s_p) :
                cmp = '' if t[2]==p else '*'
                f.write(t[0]+' '+t[1]+' cor:'+t[2]+' pred:'+p+cmp+'\n')
            f.write('\n')
#모델 평가

example_sent = test_sents[0]
print(' '.join(sent2tokens(example_sent)), end='\n\n')

print("Predicted:", ' '.join(tagger.tag(sent2features(example_sent))))
print("Correct:  ", ' '.join(sent2labels(example_sent)))

y_pred = [tagger.tag(xseq) for xseq in X_test]
write_prediction('../data/conll2002_esp.pred', test_sents, y_pred)
TP, pred_num, answer_num = eval(y_test, y_pred)
print(TP, pred_num, answer_num)

#Precision, Recall, F1score

precision = float(TP) / pred_num
recall = float(TP) / answer_num
f1score = 2.0 * (precision * recall) / (precision + recall)
print('\n\nprecision : '+str(precision*100)+'%\n'+'recall : '+str(recall*100)+'%\n'+'f1 score : '+str(f1score*100)+'%\n')



import pyTextMiner as ptm
import io
from nltk.corpus import sentiwordnet as swn
import nltk

class EnglishDictionarySentimentAnalyzer:
    def __init__(self):
        name = 'EnglishDictionarySentimentAnalyzer'

    def createDictionary(self):
        nltk.download('sentiwordnet')


if __name__ == '__main__':

    corpus = ptm.CorpusFromFile('./data/sampleEng.txt')
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.Word(),
                            ptm.helper.StopwordFilter(file='./stopwords/stopwordsEng.txt'),
                            ptm.tagger.NLTK(),
                            ptm.lemmatizer.WordNet())

    result = pipeline.processCorpus(corpus)

    EnglishDictionarySentimentAnalyzer().createDictionary()

    for doc in result:
        for sent in doc:
            for _str in sent:
                _str[0]
                _str[1]
                pos = ''
                if (str(_str[1]).startswith("N")):
                    pos = 'n'
                elif (str(_str[1]).startswith("A")):
                    pos = 'a'
                elif (str(_str[1]).startswith("V")):
                    pos = 'v'
                try:
                    if (len(pos) > 0):
                        breakdown = swn.senti_synset(str(_str[0]) + '.'+ pos + '.01')
                        print(str(breakdown) + " " + str(breakdown.pos_score())
                              + " " + str(breakdown.neg_score()) + " " + str(breakdown.obj_score()))
                except:
                    print('not found')

